"""Test Diagnostic methods"""

# pylint: disable=redefined-outer-name, no-member, too-many-public-methods
import numpy as np
import pytest

from ...data import load_arviz_data
from ...rcparams import rcParams
from ...stats import bfmi, mcse, rhat
from ...stats.diagnostics import _mc_error, ks_summary
from ...utils import Numba
from ..helpers import importorskip
from .test_diagnostics import data  # pylint: disable=unused-import

importorskip("numba")

rcParams["data.load"] = "eager"


def test_numba_bfmi():
    """Numba test for bfmi."""
    state = Numba.numba_flag
    school = load_arviz_data("centered_eight")
    data_md = np.random.rand(100, 100, 10)
    Numba.disable_numba()
    non_numba = bfmi(school.posterior["mu"].values)
    non_numba_md = bfmi(data_md)
    Numba.enable_numba()
    with_numba = bfmi(school.posterior["mu"].values)
    with_numba_md = bfmi(data_md)
    assert np.allclose(non_numba_md, with_numba_md)
    assert np.allclose(with_numba, non_numba)
    assert state == Numba.numba_flag


@pytest.mark.parametrize("method", ("rank", "split", "folded", "z_scale", "identity"))
def test_numba_rhat(method):
    """Numba test for mcse."""
    state = Numba.numba_flag
    school = np.random.rand(100, 100)
    Numba.disable_numba()
    non_numba = rhat(school, method=method)
    Numba.enable_numba()
    with_numba = rhat(school, method=method)
    assert np.allclose(with_numba, non_numba)
    assert Numba.numba_flag == state


@pytest.mark.parametrize("method", ("mean", "sd", "quantile"))
def test_numba_mcse(method, prob=None):
    """Numba test for mcse."""
    state = Numba.numba_flag
    school = np.random.rand(100, 100)
    if method == "quantile":
        prob = 0.80
    Numba.disable_numba()
    non_numba = mcse(school, method=method, prob=prob)
    Numba.enable_numba()
    with_numba = mcse(school, method=method, prob=prob)
    assert np.allclose(with_numba, non_numba)
    assert Numba.numba_flag == state


def test_ks_summary_numba():
    """Numba test for ks_summary."""
    state = Numba.numba_flag
    data = np.random.randn(100, 100)
    Numba.disable_numba()
    non_numba = (ks_summary(data)["Count"]).values
    Numba.enable_numba()
    with_numba = (ks_summary(data)["Count"]).values
    assert np.allclose(non_numba, with_numba)
    assert Numba.numba_flag == state


@pytest.mark.parametrize("batches", (1, 20))
@pytest.mark.parametrize("circular", (True, False))
def test_mcse_error_numba(batches, circular):
    """Numba test for mcse_error."""
    data = np.random.randn(100, 100)
    state = Numba.numba_flag
    Numba.disable_numba()
    non_numba = _mc_error(data, batches=batches, circular=circular)
    Numba.enable_numba()
    with_numba = _mc_error(data, batches=batches, circular=circular)
    assert np.allclose(non_numba, with_numba)
    assert state == Numba.numba_flag



# pylint: disable=too-many-lines, too-many-function-args, redefined-outer-name
"""Diagnostic functions for ArviZ."""
import warnings
from collections.abc import Sequence

import numpy as np
import packaging
import pandas as pd
import scipy
from scipy import stats

from ..data import convert_to_dataset
from ..utils import Numba, _numba_var, _stack, _var_names
from .density_utils import histogram as _histogram
from .stats_utils import _circular_standard_deviation, _sqrt
from .stats_utils import autocov as _autocov
from .stats_utils import not_valid as _not_valid
from .stats_utils import quantile as _quantile
from .stats_utils import stats_variance_2d as svar
from .stats_utils import wrap_xarray_ufunc as _wrap_xarray_ufunc

__all__ = ["bfmi", "ess", "rhat", "mcse"]


def bfmi(data):
    r"""Calculate the estimated Bayesian fraction of missing information (BFMI).

    BFMI quantifies how well momentum resampling matches the marginal energy distribution. For more
    information on BFMI, see https://arxiv.org/pdf/1604.00695v1.pdf. The current advice is that
    values smaller than 0.3 indicate poor sampling. However, this threshold is
    provisional and may change. See
    `pystan_workflow <http://mc-stan.org/users/documentation/case-studies/pystan_workflow.html>`_
    for more information.

    Parameters
    ----------
    data : obj
        Any object that can be converted to an :class:`arviz.InferenceData` object.
        Refer to documentation of :func:`arviz.convert_to_dataset` for details.
        If InferenceData, energy variable needs to be found.

    Returns
    -------
    z : array
        The Bayesian fraction of missing information of the model and trace. One element per
        chain in the trace.

    See Also
    --------
    plot_energy : Plot energy transition distribution and marginal energy
                  distribution in HMC algorithms.

    Examples
    --------
    Compute the BFMI of an InferenceData object

    .. ipython::

        In [1]: import arviz as az
           ...: data = az.load_arviz_data('radon')
           ...: az.bfmi(data)

    """
    if isinstance(data, np.ndarray):
        return _bfmi(data)

    dataset = convert_to_dataset(data, group="sample_stats")
    if not hasattr(dataset, "energy"):
        raise TypeError("Energy variable was not found.")
    return _bfmi(dataset.energy.transpose("chain", "draw"))


def ess(
    data,
    *,
    var_names=None,
    method="bulk",
    relative=False,
    prob=None,
    dask_kwargs=None,
):
    r"""Calculate estimate of the effective sample size (ess).

    Parameters
    ----------
    data : obj
        Any object that can be converted to an :class:`arviz.InferenceData` object.
        Refer to documentation of :func:`arviz.convert_to_dataset` for details.
        For ndarray: shape = (chain, draw).
        For n-dimensional ndarray transform first to dataset with :func:`arviz.convert_to_dataset`.
    var_names : str or list of str
        Names of variables to include in the return value Dataset.
    method : str, optional, default "bulk"
        Select ess method. Valid methods are:

        - "bulk"
        - "tail"     # prob, optional
        - "quantile" # prob
        - "mean" (old ess)
        - "sd"
        - "median"
        - "mad" (mean absolute deviance)
        - "z_scale"
        - "folded"
        - "identity"
        - "local"
    relative : bool
        Return relative ess
        ``ress = ess / n``
    prob : float, or tuple of two floats, optional
        probability value for "tail", "quantile" or "local" ess functions.
    dask_kwargs : dict, optional
        Dask related kwargs passed to :func:`~arviz.wrap_xarray_ufunc`.

    Returns
    -------
    xarray.Dataset
        Return the effective sample size, :math:`\hat{N}_{eff}`

    Notes
    -----
    The basic ess (:math:`N_{\mathit{eff}}`) diagnostic is computed by:

    .. math:: \hat{N}_{\mathit{eff}} = \frac{MN}{\hat{\tau}}

    .. math:: \hat{\tau} = -1 + 2 \sum_{t'=0}^K \hat{P}_{t'}

    where :math:`M` is the number of chains, :math:`N` the number of draws,
    :math:`\hat{\rho}_t` is the estimated _autocorrelation at lag :math:`t`, and
    :math:`K` is the last integer for which :math:`\hat{P}_{K} = \hat{\rho}_{2K} +
    \hat{\rho}_{2K+1}` is still positive.

    The current implementation is similar to Stan, which uses Geyer's initial monotone sequence
    criterion (Geyer, 1992; Geyer, 2011).

    References
    ----------
    * Vehtari et al. (2021). Rank-normalization, folding, and
        localization: An improved Rhat for assessing convergence of
        MCMC. Bayesian analysis, 16(2):667-718.
    * https://mc-stan.org/docs/reference-manual/analysis.html#effective-sample-size.section
    * Gelman et al. BDA3 (2013) Formula 11.8

    See Also
    --------
    arviz.rhat : Compute estimate of rank normalized splitR-hat for a set of traces.
    arviz.mcse : Calculate Markov Chain Standard Error statistic.
    plot_ess : Plot quantile, local or evolution of effective sample sizes (ESS).
    arviz.summary : Create a data frame with summary statistics.

    Examples
    --------
    Calculate the effective_sample_size using the default arguments:

    .. ipython::

        In [1]: import arviz as az
           ...: data = az.load_arviz_data('non_centered_eight')
           ...: az.ess(data)

    Calculate the ress of some of the variables

    .. ipython::

        In [1]: az.ess(data, relative=True, var_names=["mu", "theta_t"])

    Calculate the ess using the "tail" method, leaving the `prob` argument at its default
    value.

    .. ipython::

        In [1]: az.ess(data, method="tail")

    """
    methods = {
        "bulk": _ess_bulk,
        "tail": _ess_tail,
        "quantile": _ess_quantile,
        "mean": _ess_mean,
        "sd": _ess_sd,
        "median": _ess_median,
        "mad": _ess_mad,
        "z_scale": _ess_z_scale,
        "folded": _ess_folded,
        "identity": _ess_identity,
        "local": _ess_local,
    }

    if method not in methods:
        raise TypeError(f"ess method {method} not found. Valid methods are:\n{', '.join(methods)}")
    ess_func = methods[method]

    if (method == "quantile") and prob is None:
        raise TypeError("Quantile (prob) information needs to be defined.")

    if isinstance(data, np.ndarray):
        data = np.atleast_2d(data)
        if len(data.shape) < 3:
            if prob is not None:
                return ess_func(  # pylint: disable=unexpected-keyword-arg
                    data, prob=prob, relative=relative
                )

            return ess_func(data, relative=relative)

        msg = (
            "Only uni-dimensional ndarray variables are supported."
            " Please transform first to dataset with `az.convert_to_dataset`."
        )
        raise TypeError(msg)

    dataset = convert_to_dataset(data, group="posterior")
    var_names = _var_names(var_names, dataset)

    dataset = dataset if var_names is None else dataset[var_names]

    ufunc_kwargs = {"ravel": False}
    func_kwargs = {"relative": relative} if prob is None else {"prob": prob, "relative": relative}
    return _wrap_xarray_ufunc(
        ess_func,
        dataset,
        ufunc_kwargs=ufunc_kwargs,
        func_kwargs=func_kwargs,
        dask_kwargs=dask_kwargs,
    )


def rhat(data, *, var_names=None, method="rank", dask_kwargs=None):
    r"""Compute estimate of rank normalized splitR-hat for a set of traces.

    The rank normalized R-hat diagnostic tests for lack of convergence by comparing the variance
    between multiple chains to the variance within each chain. If convergence has been achieved,
    the between-chain and within-chain variances should be identical. To be most effective in
    detecting evidence for nonconvergence, each chain should have been initialized to starting
    values that are dispersed relative to the target distribution.

    Parameters
    ----------
    data : obj
        Any object that can be converted to an :class:`arviz.InferenceData` object.
        Refer to documentation of :func:`arviz.convert_to_dataset` for details.
        At least 2 posterior chains are needed to compute this diagnostic of one or more
        stochastic parameters.
        For ndarray: shape = (chain, draw).
        For n-dimensional ndarray transform first to dataset with ``az.convert_to_dataset``.
    var_names : list
        Names of variables to include in the rhat report
    method : str
        Select R-hat method. Valid methods are:
        - "rank"        # recommended by Vehtari et al. (2021)
        - "split"
        - "folded"
        - "z_scale"
        - "identity"
    dask_kwargs : dict, optional
        Dask related kwargs passed to :func:`~arviz.wrap_xarray_ufunc`.

    Returns
    -------
    xarray.Dataset
      Returns dataset of the potential scale reduction factors, :math:`\hat{R}`

    See Also
    --------
    ess : Calculate estimate of the effective sample size (ess).
    mcse : Calculate Markov Chain Standard Error statistic.
    plot_forest : Forest plot to compare HDI intervals from a number of distributions.

    Notes
    -----
    The diagnostic is computed by:

      .. math:: \hat{R} = \sqrt{\frac{\hat{V}}{W}}

    where :math:`W` is the within-chain variance and :math:`\hat{V}` is the posterior variance
    estimate for the pooled rank-traces. This is the potential scale reduction factor, which
    converges to unity when each of the traces is a sample from the target posterior. Values
    greater than one indicate that one or more chains have not yet converged.

    Rank values are calculated over all the chains with ``scipy.stats.rankdata``.
    Each chain is split in two and normalized with the z-transform following
    Vehtari et al. (2021).

    References
    ----------
    * Vehtari et al. (2021).  Rank-normalization, folding, and
      localization: An improved Rhat for assessing convergence of
      MCMC. Bayesian analysis, 16(2):667-718.
    * Gelman et al. BDA3 (2013)
    * Brooks and Gelman (1998)
    * Gelman and Rubin (1992)

    Examples
    --------
    Calculate the R-hat using the default arguments:

    .. ipython::

        In [1]: import arviz as az
           ...: data = az.load_arviz_data("non_centered_eight")
           ...: az.rhat(data)

    Calculate the R-hat of some variables using the folded method:

    .. ipython::

        In [1]: az.rhat(data, var_names=["mu", "theta_t"], method="folded")

    """
    methods = {
        "rank": _rhat_rank,
        "split": _rhat_split,
        "folded": _rhat_folded,
        "z_scale": _rhat_z_scale,
        "identity": _rhat_identity,
    }
    if method not in methods:
        raise TypeError(
            f"R-hat method {method} not found. Valid methods are:\n{', '.join(methods)}"
        )
    rhat_func = methods[method]

    if isinstance(data, np.ndarray):
        data = np.atleast_2d(data)
        if len(data.shape) < 3:
            return rhat_func(data)

        msg = (
            "Only uni-dimensional ndarray variables are supported."
            " Please transform first to dataset with `az.convert_to_dataset`."
        )
        raise TypeError(msg)

    dataset = convert_to_dataset(data, group="posterior")
    var_names = _var_names(var_names, dataset)

    dataset = dataset if var_names is None else dataset[var_names]

    ufunc_kwargs = {"ravel": False}
    func_kwargs = {}
    return _wrap_xarray_ufunc(
        rhat_func,
        dataset,
        ufunc_kwargs=ufunc_kwargs,
        func_kwargs=func_kwargs,
        dask_kwargs=dask_kwargs,
    )


def mcse(data, *, var_names=None, method="mean", prob=None, dask_kwargs=None):
    """Calculate Markov Chain Standard Error statistic.

    Parameters
    ----------
    data : obj
        Any object that can be converted to an :class:`arviz.InferenceData` object
        Refer to documentation of :func:`arviz.convert_to_dataset` for details
        For ndarray: shape = (chain, draw).
        For n-dimensional ndarray transform first to dataset with ``az.convert_to_dataset``.
    var_names : list
        Names of variables to include in the rhat report
    method : str
        Select mcse method. Valid methods are:
        - "mean"
        - "sd"
        - "median"
        - "quantile"

    prob : float
        Quantile information.
    dask_kwargs : dict, optional
        Dask related kwargs passed to :func:`~arviz.wrap_xarray_ufunc`.

    Returns
    -------
    xarray.Dataset
        Return the msce dataset

    See Also
    --------
    ess : Compute autocovariance estimates for every lag for the input array.
    summary : Create a data frame with summary statistics.
    plot_mcse : Plot quantile or local Monte Carlo Standard Error.

    Examples
    --------
    Calculate the Markov Chain Standard Error using the default arguments:

    .. ipython::

        In [1]: import arviz as az
           ...: data = az.load_arviz_data("non_centered_eight")
           ...: az.mcse(data)

    Calculate the Markov Chain Standard Error using the quantile method:

    .. ipython::

        In [1]: az.mcse(data, method="quantile", prob=0.7)

    """
    methods = {
        "mean": _mcse_mean,
        "sd": _mcse_sd,
        "median": _mcse_median,
        "quantile": _mcse_quantile,
    }
    if method not in methods:
        raise TypeError(
            "mcse method {} not found. Valid methods are:\n{}".format(
                method, "\n    ".join(methods)
            )
        )
    mcse_func = methods[method]

    if method == "quantile" and prob is None:
        raise TypeError("Quantile (prob) information needs to be defined.")

    if isinstance(data, np.ndarray):
        data = np.atleast_2d(data)
        if len(data.shape) < 3:
            if prob is not None:
                return mcse_func(data, prob=prob)  # pylint: disable=unexpected-keyword-arg

            return mcse_func(data)

        msg = (
            "Only uni-dimensional ndarray variables are supported."
            " Please transform first to dataset with `az.convert_to_dataset`."
        )
        raise TypeError(msg)

    dataset = convert_to_dataset(data, group="posterior")
    var_names = _var_names(var_names, dataset)

    dataset = dataset if var_names is None else dataset[var_names]

    ufunc_kwargs = {"ravel": False}
    func_kwargs = {} if prob is None else {"prob": prob}
    return _wrap_xarray_ufunc(
        mcse_func,
        dataset,
        ufunc_kwargs=ufunc_kwargs,
        func_kwargs=func_kwargs,
        dask_kwargs=dask_kwargs,
    )


def ks_summary(pareto_tail_indices):
    """Display a summary of Pareto tail indices.

    Parameters
    ----------
    pareto_tail_indices : array
      Pareto tail indices.

    Returns
    -------
    df_k : dataframe
      Dataframe containing k diagnostic values.
    """
    _numba_flag = Numba.numba_flag
    if _numba_flag:
        bins = np.asarray([-np.inf, 0.5, 0.7, 1, np.inf])
        kcounts, *_ = _histogram(pareto_tail_indices, bins)
    else:
        kcounts, *_ = _histogram(pareto_tail_indices, bins=[-np.inf, 0.5, 0.7, 1, np.inf])
    kprop = kcounts / len(pareto_tail_indices) * 100
    df_k = pd.DataFrame(
        dict(_=["(good)", "(ok)", "(bad)", "(very bad)"], Count=kcounts, Pct=kprop)
    ).rename(index={0: "(-Inf, 0.5]", 1: " (0.5, 0.7]", 2: "   (0.7, 1]", 3: "   (1, Inf)"})

    if np.sum(kcounts[1:]) == 0:
        warnings.warn("All Pareto k estimates are good (k < 0.5)")
    elif np.sum(kcounts[2:]) == 0:
        warnings.warn("All Pareto k estimates are ok (k < 0.7)")

    return df_k


def _bfmi(energy):
    r"""Calculate the estimated Bayesian fraction of missing information (BFMI).

    BFMI quantifies how well momentum resampling matches the marginal energy distribution. For more
    information on BFMI, see https://arxiv.org/pdf/1604.00695v1.pdf. The current advice is that
    values smaller than 0.3 indicate poor sampling. However, this threshold is provisional and may
    change. See http://mc-stan.org/users/documentation/case-studies/pystan_workflow.html for more
    information.

    Parameters
    ----------
    energy : NumPy array
        Should be extracted from a gradient based sampler, such as in Stan or PyMC3. Typically,
        after converting a trace or fit to InferenceData, the energy will be in
        `data.sample_stats.energy`.

    Returns
    -------
    z : array
        The Bayesian fraction of missing information of the model and trace. One element per
        chain in the trace.
    """
    energy_mat = np.atleast_2d(energy)
    num = np.square(np.diff(energy_mat, axis=1)).mean(axis=1)  # pylint: disable=no-member
    if energy_mat.ndim == 2:
        den = _numba_var(svar, np.var, energy_mat, axis=1, ddof=1)
    else:
        den = np.var(energy, axis=1, ddof=1)
    return num / den


def _backtransform_ranks(arr, c=3 / 8):  # pylint: disable=invalid-name
    """Backtransformation of ranks.

    Parameters
    ----------
    arr : np.ndarray
        Ranks array
    c : float
        Fractional offset. Defaults to c = 3/8 as recommended by Blom (1958).

    Returns
    -------
    np.ndarray

    References
    ----------
    Blom, G. (1958). Statistical Estimates and Transformed Beta-Variables. Wiley; New York.
    """
    arr = np.asarray(arr)
    size = arr.size
    return (arr - c) / (size - 2 * c + 1)


def _z_scale(ary):
    """Calculate z_scale.

    Parameters
    ----------
    ary : np.ndarray

    Returns
    -------
    np.ndarray
    """
    ary = np.asarray(ary)
    if packaging.version.parse(scipy.__version__) < packaging.version.parse("1.10.0.dev0"):
        rank = stats.rankdata(ary, method="average")
    else:
        # the .ravel part is only needed to overcom a bug in scipy 1.10.0.rc1
        rank = stats.rankdata(  # pylint: disable=unexpected-keyword-arg
            ary, method="average", nan_policy="omit"
        )
    rank = _backtransform_ranks(rank)
    z = stats.norm.ppf(rank)
    z = z.reshape(ary.shape)
    return z


def _split_chains(ary):
    """Split and stack chains."""
    ary = np.asarray(ary)
    if len(ary.shape) <= 1:
        ary = np.atleast_2d(ary)
    _, n_draw = ary.shape
    half = n_draw // 2
    return _stack(ary[:, :half], ary[:, -half:])


def _z_fold(ary):
    """Fold and z-scale values."""
    ary = np.asarray(ary)
    ary = abs(ary - np.median(ary))
    ary = _z_scale(ary)
    return ary


def _rhat(ary):
    """Compute the rhat for a 2d array."""
    _numba_flag = Numba.numba_flag
    ary = np.asarray(ary, dtype=float)
    if _not_valid(ary, check_shape=False):
        return np.nan
    _, num_samples = ary.shape

    # Calculate chain mean
    chain_mean = np.mean(ary, axis=1)
    # Calculate chain variance
    chain_var = _numba_var(svar, np.var, ary, axis=1, ddof=1)
    # Calculate between-chain variance
    between_chain_variance = num_samples * _numba_var(svar, np.var, chain_mean, axis=None, ddof=1)
    # Calculate within-chain variance
    within_chain_variance = np.mean(chain_var)
    # Estimate of marginal posterior variance
    rhat_value = np.sqrt(
        (between_chain_variance / within_chain_variance + num_samples - 1) / (num_samples)
    )
    return rhat_value


def _rhat_rank(ary):
    """Compute the rank normalized rhat for 2d array.

    Computation follows https://arxiv.org/abs/1903.08008
    """
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=2)):
        return np.nan
    split_ary = _split_chains(ary)
    rhat_bulk = _rhat(_z_scale(split_ary))

    split_ary_folded = abs(split_ary - np.median(split_ary))
    rhat_tail = _rhat(_z_scale(split_ary_folded))

    rhat_rank = max(rhat_bulk, rhat_tail)
    return rhat_rank


def _rhat_folded(ary):
    """Calculate split-Rhat for folded z-values."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=2)):
        return np.nan
    ary = _z_fold(_split_chains(ary))
    return _rhat(ary)


def _rhat_z_scale(ary):
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=2)):
        return np.nan
    return _rhat(_z_scale(_split_chains(ary)))


def _rhat_split(ary):
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=2)):
        return np.nan
    return _rhat(_split_chains(ary))


def _rhat_identity(ary):
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=2)):
        return np.nan
    return _rhat(ary)


def _ess(ary, relative=False):
    """Compute the effective sample size for a 2D array."""
    _numba_flag = Numba.numba_flag
    ary = np.asarray(ary, dtype=float)
    if _not_valid(ary, check_shape=False):
        return np.nan
    if (np.max(ary) - np.min(ary)) < np.finfo(float).resolution:  # pylint: disable=no-member
        return ary.size
    if len(ary.shape) < 2:
        ary = np.atleast_2d(ary)
    n_chain, n_draw = ary.shape
    acov = _autocov(ary, axis=1)
    chain_mean = ary.mean(axis=1)
    mean_var = np.mean(acov[:, 0]) * n_draw / (n_draw - 1.0)
    var_plus = mean_var * (n_draw - 1.0) / n_draw
    if n_chain > 1:
        var_plus += _numba_var(svar, np.var, chain_mean, axis=None, ddof=1)

    rho_hat_t = np.zeros(n_draw)
    rho_hat_even = 1.0
    rho_hat_t[0] = rho_hat_even
    rho_hat_odd = 1.0 - (mean_var - np.mean(acov[:, 1])) / var_plus
    rho_hat_t[1] = rho_hat_odd

    # Geyer's initial positive sequence
    t = 1
    while t < (n_draw - 3) and (rho_hat_even + rho_hat_odd) > 0.0:
        rho_hat_even = 1.0 - (mean_var - np.mean(acov[:, t + 1])) / var_plus
        rho_hat_odd = 1.0 - (mean_var - np.mean(acov[:, t + 2])) / var_plus
        if (rho_hat_even + rho_hat_odd) >= 0:
            rho_hat_t[t + 1] = rho_hat_even
            rho_hat_t[t + 2] = rho_hat_odd
        t += 2

    max_t = t - 2
    # improve estimation
    if rho_hat_even > 0:
        rho_hat_t[max_t + 1] = rho_hat_even
    # Geyer's initial monotone sequence
    t = 1
    while t <= max_t - 2:
        if (rho_hat_t[t + 1] + rho_hat_t[t + 2]) > (rho_hat_t[t - 1] + rho_hat_t[t]):
            rho_hat_t[t + 1] = (rho_hat_t[t - 1] + rho_hat_t[t]) / 2.0
            rho_hat_t[t + 2] = rho_hat_t[t + 1]
        t += 2

    ess = n_chain * n_draw
    tau_hat = -1.0 + 2.0 * np.sum(rho_hat_t[: max_t + 1]) + np.sum(rho_hat_t[max_t + 1 : max_t + 2])
    tau_hat = max(tau_hat, 1 / np.log10(ess))
    ess = (1 if relative else ess) / tau_hat
    if np.isnan(rho_hat_t).any():
        ess = np.nan
    return ess


def _ess_bulk(ary, relative=False):
    """Compute the effective sample size for the bulk."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    z_scaled = _z_scale(_split_chains(ary))
    ess_bulk = _ess(z_scaled, relative=relative)
    return ess_bulk


def _ess_tail(ary, prob=None, relative=False):
    """Compute the effective sample size for the tail.

    If `prob` defined, ess = min(qess(prob), qess(1-prob))
    """
    if prob is None:
        prob = (0.05, 0.95)
    elif not isinstance(prob, Sequence):
        prob = (prob, 1 - prob)

    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan

    prob_low, prob_high = prob
    quantile_low_ess = _ess_quantile(ary, prob_low, relative=relative)
    quantile_high_ess = _ess_quantile(ary, prob_high, relative=relative)
    return min(quantile_low_ess, quantile_high_ess)


def _ess_mean(ary, relative=False):
    """Compute the effective sample size for the mean."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    return _ess(_split_chains(ary), relative=relative)


def _ess_sd(ary, relative=False):
    """Compute the effective sample size for the sd."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    ary = (ary - ary.mean()) ** 2
    return _ess(_split_chains(ary), relative=relative)


def _ess_quantile(ary, prob, relative=False):
    """Compute the effective sample size for the specific residual."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    if prob is None:
        raise TypeError("Prob not defined.")
    (quantile,) = _quantile(ary, prob)
    iquantile = ary <= quantile
    return _ess(_split_chains(iquantile), relative=relative)


def _ess_local(ary, prob, relative=False):
    """Compute the effective sample size for the specific residual."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    if prob is None:
        raise TypeError("Prob not defined.")
    if len(prob) != 2:
        raise ValueError("Prob argument in ess local must be upper and lower bound")
    quantile = _quantile(ary, prob)
    iquantile = (quantile[0] <= ary) & (ary <= quantile[1])
    return _ess(_split_chains(iquantile), relative=relative)


def _ess_z_scale(ary, relative=False):
    """Calculate ess for z-scaLe."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    return _ess(_z_scale(_split_chains(ary)), relative=relative)


def _ess_folded(ary, relative=False):
    """Calculate split-ess for folded data."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    return _ess(_z_fold(_split_chains(ary)), relative=relative)


def _ess_median(ary, relative=False):
    """Calculate split-ess for median."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    return _ess_quantile(ary, 0.5, relative=relative)


def _ess_mad(ary, relative=False):
    """Calculate split-ess for mean absolute deviance."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    ary = abs(ary - np.median(ary))
    ary = ary <= np.median(ary)
    ary = _z_scale(_split_chains(ary))
    return _ess(ary, relative=relative)


def _ess_identity(ary, relative=False):
    """Calculate ess."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    return _ess(ary, relative=relative)


def _mcse_mean(ary):
    """Compute the Markov Chain mean error."""
    _numba_flag = Numba.numba_flag
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    ess = _ess_mean(ary)
    if _numba_flag:
        sd = _sqrt(svar(np.ravel(ary), ddof=1), np.zeros(1))
    else:
        sd = np.std(ary, ddof=1)
    mcse_mean_value = sd / np.sqrt(ess)
    return mcse_mean_value


def _mcse_sd(ary):
    """Compute the Markov Chain sd error."""
    _numba_flag = Numba.numba_flag
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    sims_c2 = (ary - ary.mean()) ** 2
    ess = _ess_mean(sims_c2)
    evar = (sims_c2).mean()
    varvar = ((sims_c2**2).mean() - evar**2) / ess
    varsd = varvar / evar / 4
    if _numba_flag:
        mcse_sd_value = float(_sqrt(np.ravel(varsd), np.zeros(1)))
    else:
        mcse_sd_value = np.sqrt(varsd)
    return mcse_sd_value


def _mcse_median(ary):
    """Compute the Markov Chain median error."""
    return _mcse_quantile(ary, 0.5)


def _mcse_quantile(ary, prob):
    """Compute the Markov Chain quantile error at quantile=prob."""
    ary = np.asarray(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        return np.nan
    ess = _ess_quantile(ary, prob)
    probability = [0.1586553, 0.8413447]
    with np.errstate(invalid="ignore"):
        ppf = stats.beta.ppf(probability, ess * prob + 1, ess * (1 - prob) + 1)
    sorted_ary = np.sort(ary.ravel())
    size = sorted_ary.size
    ppf_size = ppf * size - 1
    th1 = sorted_ary[int(np.floor(np.nanmax((ppf_size[0], 0))))]
    th2 = sorted_ary[int(np.ceil(np.nanmin((ppf_size[1], size - 1))))]
    return (th2 - th1) / 2


def _mc_error(ary, batches=5, circular=False):
    """Calculate the simulation standard error, accounting for non-independent samples.

    The trace is divided into batches, and the standard deviation of the batch
    means is calculated.

    Parameters
    ----------
    ary : Numpy array
        An array containing MCMC samples
    batches : integer
        Number of batches
    circular : bool
        Whether to compute the error taking into account `ary` is a circular variable
        (in the range [-np.pi, np.pi]) or not. Defaults to False (i.e non-circular variables).

    Returns
    -------
    mc_error : float
        Simulation standard error
    """
    _numba_flag = Numba.numba_flag
    if ary.ndim > 1:
        dims = np.shape(ary)
        trace = np.transpose([t.ravel() for t in ary])

        return np.reshape([_mc_error(t, batches) for t in trace], dims[1:])

    else:
        if _not_valid(ary, check_shape=False):
            return np.nan
        if batches == 1:
            if circular:
                if _numba_flag:
                    std = _circular_standard_deviation(ary, high=np.pi, low=-np.pi)
                else:
                    std = stats.circstd(ary, high=np.pi, low=-np.pi)
            elif _numba_flag:
                std = float(_sqrt(svar(ary), np.zeros(1)).item())
            else:
                std = np.std(ary)
            return std / np.sqrt(len(ary))

        batched_traces = np.resize(ary, (batches, int(len(ary) / batches)))

        if circular:
            means = stats.circmean(batched_traces, high=np.pi, low=-np.pi, axis=1)
            if _numba_flag:
                std = _circular_standard_deviation(means, high=np.pi, low=-np.pi)
            else:
                std = stats.circstd(means, high=np.pi, low=-np.pi)
        else:
            means = np.mean(batched_traces, 1)
            std = _sqrt(svar(means), np.zeros(1)) if _numba_flag else np.std(means)
        return std / np.sqrt(batches)


def _multichain_statistics(ary, focus="mean"):
    """Calculate efficiently multichain statistics for summary.

    Parameters
    ----------
    ary : numpy.ndarray
    focus : select focus for the statistics. Deafault is mean.

    Returns
    -------
    tuple
        Order of return parameters is
            If focus equals "mean"
                - mcse_mean, mcse_sd, ess_bulk, ess_tail, r_hat
            Else if focus equals "median"
                - mcse_median, ess_median, ess_tail, r_hat
    """
    ary = np.atleast_2d(ary)
    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=1)):
        if focus == "mean":
            return np.nan, np.nan, np.nan, np.nan, np.nan
        return np.nan, np.nan, np.nan, np.nan

    z_split = _z_scale(_split_chains(ary))

    # ess tail
    quantile05, quantile95 = _quantile(ary, [0.05, 0.95])
    iquantile05 = ary <= quantile05
    quantile05_ess = _ess(_split_chains(iquantile05))
    iquantile95 = ary <= quantile95
    quantile95_ess = _ess(_split_chains(iquantile95))
    ess_tail_value = min(quantile05_ess, quantile95_ess)

    if _not_valid(ary, shape_kwargs=dict(min_draws=4, min_chains=2)):
        rhat_value = np.nan
    else:
        # r_hat
        rhat_bulk = _rhat(z_split)
        ary_folded = np.abs(ary - np.median(ary))
        rhat_tail = _rhat(_z_scale(_split_chains(ary_folded)))
        rhat_value = max(rhat_bulk, rhat_tail)

    if focus == "mean":
        # ess mean
        ess_mean_value = _ess_mean(ary)

        # mcse_mean
        sims_c2 = (ary - ary.mean()) ** 2
        sims_c2_sum = sims_c2.sum()
        var = sims_c2_sum / (sims_c2.size - 1)
        mcse_mean_value = np.sqrt(var / ess_mean_value)

        # ess bulk
        ess_bulk_value = _ess(z_split)

        # mcse_sd
        evar = sims_c2_sum / sims_c2.size
        ess_mean_sims = _ess_mean(sims_c2)
        varvar = ((sims_c2**2).mean() - evar**2) / ess_mean_sims
        varsd = varvar / evar / 4
        mcse_sd_value = np.sqrt(varsd)

        return (
            mcse_mean_value,
            mcse_sd_value,
            ess_bulk_value,
            ess_tail_value,
            rhat_value,
        )

    # ess median
    ess_median_value = _ess_median(ary)

    # mcse_median
    mcse_median_value = _mcse_median(ary)

    return (
        mcse_median_value,
        ess_median_value,
        ess_tail_value,
        rhat_value,
    )



"""Matplotlib hdiplot."""

import matplotlib.pyplot as plt
from matplotlib import _pylab_helpers

from ...plot_utils import _scale_fig_size, vectorized_to_hex
from . import backend_kwarg_defaults, backend_show, create_axes_grid, matplotlib_kwarg_dealiaser


def plot_hdi(ax, x_data, y_data, color, figsize, plot_kwargs, fill_kwargs, backend_kwargs, show):
    """Matplotlib HDI plot."""
    if backend_kwargs is None:
        backend_kwargs = {}

    backend_kwargs = {
        **backend_kwarg_defaults(),
        **backend_kwargs,
    }

    plot_kwargs = matplotlib_kwarg_dealiaser(plot_kwargs, "plot")
    plot_kwargs["color"] = vectorized_to_hex(plot_kwargs.get("color", color))
    plot_kwargs.setdefault("alpha", 0)

    fill_kwargs = matplotlib_kwarg_dealiaser(fill_kwargs, "fill_between")
    fill_kwargs["color"] = vectorized_to_hex(fill_kwargs.get("color", color))
    fill_kwargs.setdefault("alpha", 0.5)

    figsize, *_ = _scale_fig_size(figsize, None)
    backend_kwargs.setdefault("figsize", figsize)
    backend_kwargs["squeeze"] = True

    if ax is None:
        fig_manager = _pylab_helpers.Gcf.get_active()
        if fig_manager is not None:
            ax = fig_manager.canvas.figure.gca()
        else:
            _, ax = create_axes_grid(
                1,
                backend_kwargs=backend_kwargs,
            )

    ax.plot(x_data, y_data, **plot_kwargs)
    ax.fill_between(x_data, y_data[:, 0], y_data[:, 1], **fill_kwargs)

    if backend_show(show):
        plt.show()

    return ax



"""Bokeh Traceplot."""

import warnings
from collections.abc import Iterable
from itertools import cycle

import bokeh.plotting as bkp
import matplotlib.pyplot as plt
import numpy as np
from bokeh.models import ColumnDataSource, DataRange1d, Span
from bokeh.models.glyphs import Scatter
from bokeh.models.annotations import Title

from ...distplot import plot_dist
from ...plot_utils import _scale_fig_size
from ...rankplot import plot_rank
from .. import show_layout
from . import backend_kwarg_defaults, dealiase_sel_kwargs
from ....sel_utils import xarray_var_iter


def plot_trace(
    data,
    var_names,
    divergences,
    kind,
    figsize,
    rug,
    lines,
    circ_var_names,  # pylint: disable=unused-argument
    circ_var_units,  # pylint: disable=unused-argument
    compact,
    compact_prop,
    combined,
    chain_prop,
    legend,
    labeller,
    plot_kwargs,
    fill_kwargs,
    rug_kwargs,
    hist_kwargs,
    trace_kwargs,
    rank_kwargs,
    plotters,
    divergence_data,
    axes,
    backend_kwargs,
    backend_config,
    show,
):
    """Bokeh traceplot."""
    # If divergences are plotted they must be provided
    if divergences is not False:
        assert divergence_data is not None

    if backend_config is None:
        backend_config = {}

    backend_config = {
        **backend_kwarg_defaults(
            ("bounds_y_range", "plot.bokeh.bounds_y_range"),
        ),
        **backend_config,
    }

    # Set plot default backend kwargs
    if backend_kwargs is None:
        backend_kwargs = {}

    backend_kwargs = {
        **backend_kwarg_defaults(
            ("dpi", "plot.bokeh.figure.dpi"),
        ),
        **backend_kwargs,
    }
    dpi = backend_kwargs.pop("dpi")

    if figsize is None:
        figsize = (12, len(plotters) * 2)

    figsize, _, _, _, linewidth, _ = _scale_fig_size(figsize, 10, rows=len(plotters), cols=2)

    backend_kwargs.setdefault("height", int(figsize[1] * dpi // len(plotters)))
    backend_kwargs.setdefault("width", int(figsize[0] * dpi // 2))

    if lines is None:
        lines = ()

    num_chain_props = len(data.chain) + 1 if combined else len(data.chain)
    if not compact:
        chain_prop = (
            {"line_color": plt.rcParams["axes.prop_cycle"].by_key()["color"]}
            if chain_prop is None
            else chain_prop
        )
    else:
        chain_prop = (
            {
                "line_dash": ("solid", "dotted", "dashed", "dashdot"),
            }
            if chain_prop is None
            else chain_prop
        )
        compact_prop = (
            {"line_color": plt.rcParams["axes.prop_cycle"].by_key()["color"]}
            if compact_prop is None
            else compact_prop
        )

    if isinstance(chain_prop, str):
        chain_prop = {chain_prop: plt.rcParams["axes.prop_cycle"].by_key()[chain_prop]}
    if isinstance(chain_prop, tuple):
        warnings.warn(
            "chain_prop as a tuple will be deprecated in a future warning, use a dict instead",
            FutureWarning,
        )
        chain_prop = {chain_prop[0]: chain_prop[1]}
    chain_prop = {
        prop_name: [prop for _, prop in zip(range(num_chain_props), cycle(props))]
        for prop_name, props in chain_prop.items()
    }

    if isinstance(compact_prop, str):
        compact_prop = {compact_prop: plt.rcParams["axes.prop_cycle"].by_key()[compact_prop]}
    if isinstance(compact_prop, tuple):
        warnings.warn(
            "compact_prop as a tuple will be deprecated in a future warning, use a dict instead",
            FutureWarning,
        )
        compact_prop = {compact_prop[0]: compact_prop[1]}

    trace_kwargs = {} if trace_kwargs is None else trace_kwargs
    trace_kwargs.setdefault("alpha", 0.35)

    if hist_kwargs is None:
        hist_kwargs = {}
    hist_kwargs.setdefault("alpha", 0.35)

    if plot_kwargs is None:
        plot_kwargs = {}
    if fill_kwargs is None:
        fill_kwargs = {}
    if rug_kwargs is None:
        rug_kwargs = {}
    if rank_kwargs is None:
        rank_kwargs = {}

    trace_kwargs.setdefault("line_width", linewidth)
    plot_kwargs.setdefault("line_width", linewidth)

    if rank_kwargs is None:
        rank_kwargs = {}

    if axes is None:
        axes = []
        backend_kwargs_copy = backend_kwargs.copy()
        for i in range(len(plotters)):
            if not i:
                _axes = [bkp.figure(**backend_kwargs), bkp.figure(**backend_kwargs_copy)]
                backend_kwargs_copy.setdefault("x_range", _axes[1].x_range)
            else:
                _axes = [
                    bkp.figure(**backend_kwargs),
                    bkp.figure(**backend_kwargs_copy),
                ]
            axes.append(_axes)

    axes = np.atleast_2d(axes)

    cds_data = {}
    cds_var_groups = {}
    draw_name = "draw"

    for var_name, selection, isel, value in list(
        xarray_var_iter(data, var_names=var_names, combined=True)
    ):
        if selection:
            cds_name = "{}_ARVIZ_CDS_SELECTION_{}".format(
                var_name,
                "_".join(
                    str(item)
                    for key, value in selection.items()
                    for item in (
                        [key, value]
                        if (isinstance(value, str) or not isinstance(value, Iterable))
                        else [key, *value]
                    )
                ),
            )
        else:
            cds_name = var_name

        if var_name not in cds_var_groups:
            cds_var_groups[var_name] = []
        cds_var_groups[var_name].append(cds_name)

        for chain_idx, _ in enumerate(data.chain.values):
            if chain_idx not in cds_data:
                cds_data[chain_idx] = {}
            _data = value[chain_idx]
            cds_data[chain_idx][cds_name] = _data

    while any(key == draw_name for key in cds_data[0]):
        draw_name += "w"

    for chain in cds_data.values():
        chain[draw_name] = data.draw.values

    cds_data = {chain_idx: ColumnDataSource(cds) for chain_idx, cds in cds_data.items()}

    for idx, (var_name, selection, isel, value) in enumerate(plotters):
        value = np.atleast_2d(value)

        if len(value.shape) == 2:
            y_name = (
                var_name
                if not selection
                else "{}_ARVIZ_CDS_SELECTION_{}".format(
                    var_name,
                    "_".join(
                        str(item)
                        for key, value in selection.items()
                        for item in (
                            (key, value)
                            if (isinstance(value, str) or not isinstance(value, Iterable))
                            else (key, *value)
                        )
                    ),
                )
            )
            if rug:
                rug_kwargs["y"] = y_name
            _plot_chains_bokeh(
                ax_density=axes[idx, 0],
                ax_trace=axes[idx, 1],
                data=cds_data,
                x_name=draw_name,
                y_name=y_name,
                chain_prop=chain_prop,
                combined=combined,
                rug=rug,
                kind=kind,
                legend=legend,
                trace_kwargs=trace_kwargs,
                hist_kwargs=hist_kwargs,
                plot_kwargs=plot_kwargs,
                fill_kwargs=fill_kwargs,
                rug_kwargs=rug_kwargs,
                rank_kwargs=rank_kwargs,
            )
        else:
            for y_name in cds_var_groups[var_name]:
                if rug:
                    rug_kwargs["y"] = y_name
                _plot_chains_bokeh(
                    ax_density=axes[idx, 0],
                    ax_trace=axes[idx, 1],
                    data=cds_data,
                    x_name=draw_name,
                    y_name=y_name,
                    chain_prop=chain_prop,
                    combined=combined,
                    rug=rug,
                    kind=kind,
                    legend=legend,
                    trace_kwargs=trace_kwargs,
                    hist_kwargs=hist_kwargs,
                    plot_kwargs=plot_kwargs,
                    fill_kwargs=fill_kwargs,
                    rug_kwargs=rug_kwargs,
                    rank_kwargs=rank_kwargs,
                )

        for col in (0, 1):
            _title = Title()
            _title.text = labeller.make_label_vert(var_name, selection, isel)
            axes[idx, col].title = _title
            axes[idx, col].y_range = DataRange1d(
                bounds=backend_config["bounds_y_range"], min_interval=0.1
            )

        for _, _, vlines in (j for j in lines if j[0] == var_name and j[1] == selection):
            if isinstance(vlines, (float, int)):
                line_values = [vlines]
            else:
                line_values = np.atleast_1d(vlines).ravel()

            for line_value in line_values:
                vline = Span(
                    location=line_value,
                    dimension="height",
                    line_color="black",
                    line_width=1.5,
                    line_alpha=0.75,
                )
                hline = Span(
                    location=line_value,
                    dimension="width",
                    line_color="black",
                    line_width=1.5,
                    line_alpha=trace_kwargs["alpha"],
                )

                axes[idx, 0].renderers.append(vline)
                axes[idx, 1].renderers.append(hline)

        if legend:
            for col in (0, 1):
                axes[idx, col].legend.location = "top_left"
                axes[idx, col].legend.click_policy = "hide"
        else:
            for col in (0, 1):
                if axes[idx, col].legend:
                    axes[idx, col].legend.visible = False

        if divergences:
            div_density_kwargs = {}
            div_density_kwargs.setdefault("size", 14)
            div_density_kwargs.setdefault("line_color", "red")
            div_density_kwargs.setdefault("line_width", 2)
            div_density_kwargs.setdefault("line_alpha", 0.50)
            div_density_kwargs.setdefault("angle", np.pi / 2)

            div_trace_kwargs = {}
            div_trace_kwargs.setdefault("size", 14)
            div_trace_kwargs.setdefault("line_color", "red")
            div_trace_kwargs.setdefault("line_width", 2)
            div_trace_kwargs.setdefault("line_alpha", 0.50)
            div_trace_kwargs.setdefault("angle", np.pi / 2)

            div_selection = {k: v for k, v in selection.items() if k in divergence_data.dims}
            divs = divergence_data.sel(**div_selection).values
            divs = np.atleast_2d(divs)

            for chain, chain_divs in enumerate(divs):
                div_idxs = np.arange(len(chain_divs))[chain_divs]
                if div_idxs.size > 0:
                    values = value[chain, div_idxs]
                    tmp_cds = ColumnDataSource({"y": values, "x": div_idxs})
                    if divergences == "top":
                        y_div_trace = value.max()
                    else:
                        y_div_trace = value.min()
                    glyph_density = Scatter(x="y", y=0.0, marker="dash", **div_density_kwargs)
                    if kind == "trace":
                        glyph_trace = Scatter(
                            x="x", y=y_div_trace, marker="dash", **div_trace_kwargs
                        )
                        axes[idx, 1].add_glyph(tmp_cds, glyph_trace)

                    axes[idx, 0].add_glyph(tmp_cds, glyph_density)

    show_layout(axes, show)

    return axes


def _plot_chains_bokeh(
    ax_density,
    ax_trace,
    data,
    x_name,
    y_name,
    chain_prop,
    combined,
    rug,
    kind,
    legend,
    trace_kwargs,
    hist_kwargs,
    plot_kwargs,
    fill_kwargs,
    rug_kwargs,
    rank_kwargs,
):
    marker = trace_kwargs.pop("marker", True)
    for chain_idx, cds in data.items():
        if kind == "trace":
            if legend:
                trace_kwargs["legend_label"] = f"chain {chain_idx}"
            ax_trace.line(
                x=x_name,
                y=y_name,
                source=cds,
                **dealiase_sel_kwargs(trace_kwargs, chain_prop, chain_idx),
            )
            if marker:
                ax_trace.scatter(
                    x=x_name,
                    y=y_name,
                    marker="circle",
                    source=cds,
                    radius=0.30,
                    alpha=0.5,
                    **dealiase_sel_kwargs({}, chain_prop, chain_idx),
                )
        if not combined:
            rug_kwargs["cds"] = cds
            if legend:
                plot_kwargs["legend_label"] = f"chain {chain_idx}"
            plot_dist(
                cds.data[y_name],
                ax=ax_density,
                rug=rug,
                hist_kwargs=hist_kwargs,
                plot_kwargs=dealiase_sel_kwargs(plot_kwargs, chain_prop, chain_idx),
                fill_kwargs=fill_kwargs,
                rug_kwargs=rug_kwargs,
                backend="bokeh",
                backend_kwargs={},
                show=False,
            )

    if kind == "rank_bars":
        value = np.array([item.data[y_name] for item in data.values()])
        plot_rank(value, kind="bars", ax=ax_trace, backend="bokeh", show=False, **rank_kwargs)
    elif kind == "rank_vlines":
        value = np.array([item.data[y_name] for item in data.values()])
        plot_rank(value, kind="vlines", ax=ax_trace, backend="bokeh", show=False, **rank_kwargs)

    if combined:
        rug_kwargs["cds"] = data
        if legend:
            plot_kwargs["legend_label"] = "combined chains"
        plot_dist(
            np.concatenate([item.data[y_name] for item in data.values()]).flatten(),
            ax=ax_density,
            rug=rug,
            hist_kwargs=hist_kwargs,
            plot_kwargs=dealiase_sel_kwargs(plot_kwargs, chain_prop, -1),
            fill_kwargs=fill_kwargs,
            rug_kwargs=rug_kwargs,
            backend="bokeh",
            backend_kwargs={},
            show=False,
        )



"""Plot quantile or local effective sample sizes."""

import numpy as np
import xarray as xr

from ..data import convert_to_dataset
from ..labels import BaseLabeller
from ..rcparams import rcParams
from ..sel_utils import xarray_var_iter
from ..stats import ess
from ..utils import _var_names, get_coords
from .plot_utils import default_grid, filter_plotters_list, get_plotting_function


def plot_ess(
    idata,
    var_names=None,
    filter_vars=None,
    kind="local",
    relative=False,
    coords=None,
    figsize=None,
    grid=None,
    textsize=None,
    rug=False,
    rug_kind="diverging",
    n_points=20,
    extra_methods=False,
    min_ess=400,
    labeller=None,
    ax=None,
    extra_kwargs=None,
    text_kwargs=None,
    hline_kwargs=None,
    rug_kwargs=None,
    backend=None,
    backend_kwargs=None,
    show=None,
    **kwargs,
):
    r"""Generate quantile, local, or evolution ESS plots.

    The local and the quantile ESS plots are recommended for checking
    that there are enough samples for all the explored regions of the
    parameter space. Checking local and quantile ESS is particularly
    relevant when working with HDI intervals as opposed to ESS bulk,
    which is suitable for point estimates.

    Parameters
    ----------
    idata : InferenceData
        Any object that can be converted to an :class:`arviz.InferenceData` object
        Refer to documentation of :func:`arviz.convert_to_dataset` for details.
    var_names : list of str, optional
        Variables to be plotted. Prefix the variables by ``~`` when you want to exclude
        them from the plot. See :ref:`this section <common_var_names>` for usage examples.
    filter_vars : {None, "like", "regex"}, default None
        If `None` (default), interpret `var_names` as the real variables names. If "like",
        interpret `var_names` as substrings of the real variables names. If "regex",
        interpret `var_names` as regular expressions on the real variables names. See
        :ref:`this section <common_filter_vars>` for usage examples.
    kind : {"local", "quantile", "evolution"}, default "local"
        Specify the kind of plot:

        * The ``kind="local"`` argument generates the ESS' local efficiency for
          estimating quantiles of a desired posterior.
        * The ``kind="quantile"`` argument generates the ESS' local efficiency
          for estimating small-interval probability of a desired posterior.
        * The ``kind="evolution"`` argument generates the estimated ESS'
          with incrised number of iterations of a desired posterior.

    relative : bool, default False
        Show relative ess in plot ``ress = ess / N``.
    coords : dict, optional
        Coordinates of `var_names` to be plotted. Passed to :meth:`xarray.Dataset.sel`.
        See :ref:`this section <common_coords>` for usage examples.
    grid : tuple, optional
        Number of rows and columns. By default, the rows and columns are
        automatically inferred. See :ref:`this section <common_grid>` for usage examples.
    figsize : (float, float), optional
        Figure size. If ``None`` it will be defined automatically.
    textsize : float, optional
        Text size scaling factor for labels, titles and lines. If ``None`` it will be autoscaled
        based on `figsize`.
    rug : bool, default False
        Add a `rug plot <https://en.wikipedia.org/wiki/Rug_plot>`_ for a specific subset of values.
    rug_kind : str, default "diverging"
        Variable in sample stats to use as rug mask. Must be a boolean variable.
    n_points : int, default 20
        Number of points for which to plot their quantile/local ess or number of subsets
        in the evolution plot.
    extra_methods : bool, default False
        Plot mean and sd ESS as horizontal lines. Not taken into account if ``kind = 'evolution'``.
    min_ess : int, default 400
        Minimum number of ESS desired. If ``relative=True`` the line is plotted at
        ``min_ess / n_samples`` for local and quantile kinds and as a curve following
        the ``min_ess / n`` dependency in evolution kind.
    labeller : Labeller, optional
        Class providing the method ``make_label_vert`` to generate the labels in the plot titles.
        Read the :ref:`label_guide` for more details and usage examples.
    ax : 2D array-like of matplotlib_axes or bokeh_figure, optional
        A 2D array of locations into which to plot the densities. If not supplied, ArviZ will create
        its own array of plot areas (and return it).
    extra_kwargs : dict, optional
        If evolution plot, `extra_kwargs` is used to plot ess tail and differentiate it
        from ess bulk. Otherwise, passed to extra methods lines.
    text_kwargs : dict, optional
        Only taken into account when ``extra_methods=True``. kwargs passed to ax.annotate
        for extra methods lines labels. It accepts the additional
        key ``x`` to set ``xy=(text_kwargs["x"], mcse)``
    hline_kwargs : dict, optional
        kwargs passed to :func:`~matplotlib.axes.Axes.axhline` or to :class:`~bokeh.models.Span`
        depending on the backend for the horizontal minimum ESS line.
        For relative ess evolution plots the kwargs are passed to
        :func:`~matplotlib.axes.Axes.plot` or to :class:`~bokeh.plotting.figure.line`
    rug_kwargs : dict
        kwargs passed to rug plot.
    backend : {"matplotlib", "bokeh"}, default "matplotlib"
        Select plotting backend.
    backend_kwargs : dict, optional
        These are kwargs specific to the backend being used, passed to
        :func:`matplotlib.pyplot.subplots` or :class:`bokeh.plotting.figure`.
        For additional documentation check the plotting method of the backend.
    show : bool, optional
        Call backend show function.
    **kwargs
        Passed as-is to :meth:`mpl:matplotlib.axes.Axes.hist` or
        :meth:`mpl:matplotlib.axes.Axes.plot` function depending on the
        value of `kind`.

    Returns
    -------
    axes : matplotlib_axes or bokeh_figure

    See Also
    --------
    ess : Calculate estimate of the effective sample size.

    References
    ----------
    .. [1] Vehtari et al. (2021). Rank-normalization, folding, and
        localization: An improved Rhat for assessing convergence of
        MCMC. Bayesian analysis, 16(2):667-718.

    Examples
    --------
    Plot local ESS.

    .. plot::
        :context: close-figs

        >>> import arviz as az
        >>> idata = az.load_arviz_data("centered_eight")
        >>> coords = {"school": ["Choate", "Lawrenceville"]}
        >>> az.plot_ess(
        ...     idata, kind="local", var_names=["mu", "theta"], coords=coords
        ... )

    Plot ESS evolution as the number of samples increase. When the model is converging properly,
    both lines in this plot should be roughly linear.

    .. plot::
        :context: close-figs

        >>> az.plot_ess(
        ...     idata, kind="evolution", var_names=["mu", "theta"], coords=coords
        ... )

    Customize local ESS plot to look like reference paper.

    .. plot::
        :context: close-figs

        >>> az.plot_ess(
        ...     idata, kind="local", var_names=["mu"], drawstyle="steps-mid", color="k",
        ...     linestyle="-", marker=None, rug=True, rug_kwargs={"color": "r"}
        ... )

    Customize ESS evolution plot to look like reference paper.

    .. plot::
        :context: close-figs

        >>> extra_kwargs = {"color": "lightsteelblue"}
        >>> az.plot_ess(
        ...     idata, kind="evolution", var_names=["mu"],
        ...     color="royalblue", extra_kwargs=extra_kwargs
        ... )

    """
    valid_kinds = ("local", "quantile", "evolution")
    kind = kind.lower()
    if kind not in valid_kinds:
        raise ValueError(f"Invalid kind, kind must be one of {valid_kinds} not {kind}")

    if coords is None:
        coords = {}
    if "chain" in coords or "draw" in coords:
        raise ValueError("chain and draw are invalid coordinates for this kind of plot")
    if labeller is None:
        labeller = BaseLabeller()
    extra_methods = False if kind == "evolution" else extra_methods

    data = get_coords(convert_to_dataset(idata, group="posterior"), coords)
    var_names = _var_names(var_names, data, filter_vars)
    n_draws = data.sizes["draw"]
    n_samples = n_draws * data.sizes["chain"]

    ess_tail_dataset = None
    mean_ess = None
    sd_ess = None

    if kind == "quantile":
        probs = np.linspace(1 / n_points, 1 - 1 / n_points, n_points)
        xdata = probs
        ylabel = "{} for quantiles"
        ess_dataset = xr.concat(
            [
                ess(data, var_names=var_names, relative=relative, method="quantile", prob=p)
                for p in probs
            ],
            dim="ess_dim",
        )
    elif kind == "local":
        probs = np.linspace(0, 1, n_points, endpoint=False)
        xdata = probs
        ylabel = "{} for small intervals"
        ess_dataset = xr.concat(
            [
                ess(
                    data,
                    var_names=var_names,
                    relative=relative,
                    method="local",
                    prob=[p, p + 1 / n_points],
                )
                for p in probs
            ],
            dim="ess_dim",
        )
    else:
        first_draw = data.draw.values[0]
        ylabel = "{}"
        xdata = np.linspace(n_samples / n_points, n_samples, n_points)
        draw_divisions = np.linspace(n_draws // n_points, n_draws, n_points, dtype=int)
        ess_dataset = xr.concat(
            [
                ess(
                    data.sel(draw=slice(first_draw + draw_div)),
                    var_names=var_names,
                    relative=relative,
                    method="bulk",
                )
                for draw_div in draw_divisions
            ],
            dim="ess_dim",
        )
        ess_tail_dataset = xr.concat(
            [
                ess(
                    data.sel(draw=slice(first_draw + draw_div)),
                    var_names=var_names,
                    relative=relative,
                    method="tail",
                )
                for draw_div in draw_divisions
            ],
            dim="ess_dim",
        )

    plotters = filter_plotters_list(
        list(xarray_var_iter(ess_dataset, var_names=var_names, skip_dims={"ess_dim"})), "plot_ess"
    )
    length_plotters = len(plotters)
    rows, cols = default_grid(length_plotters, grid=grid)

    if extra_methods:
        mean_ess = ess(data, var_names=var_names, method="mean", relative=relative)
        sd_ess = ess(data, var_names=var_names, method="sd", relative=relative)

    essplot_kwargs = dict(
        ax=ax,
        plotters=plotters,
        xdata=xdata,
        ess_tail_dataset=ess_tail_dataset,
        mean_ess=mean_ess,
        sd_ess=sd_ess,
        idata=idata,
        data=data,
        kind=kind,
        extra_methods=extra_methods,
        textsize=textsize,
        rows=rows,
        cols=cols,
        figsize=figsize,
        kwargs=kwargs,
        extra_kwargs=extra_kwargs,
        text_kwargs=text_kwargs,
        n_samples=n_samples,
        relative=relative,
        min_ess=min_ess,
        labeller=labeller,
        ylabel=ylabel,
        rug=rug,
        rug_kind=rug_kind,
        rug_kwargs=rug_kwargs,
        hline_kwargs=hline_kwargs,
        backend_kwargs=backend_kwargs,
        show=show,
    )

    if backend is None:
        backend = rcParams["plot.backend"]
    backend = backend.lower()

    # TODO: Add backend kwargs
    plot = get_plotting_function("plot_ess", "essplot", backend)
    ax = plot(**essplot_kwargs)
    return ax



"""Matplotlib ecdfplot."""

import matplotlib.pyplot as plt
from matplotlib.colors import to_hex

from ...plot_utils import _scale_fig_size
from . import backend_kwarg_defaults, create_axes_grid, backend_show


def plot_ecdf(
    x_coord,
    y_coord,
    x_bands,
    lower,
    higher,
    plot_kwargs,
    fill_kwargs,
    plot_outline_kwargs,
    figsize,
    fill_band,
    ax,
    show,
    backend_kwargs,
):
    """Matplotlib ecdfplot."""
    if backend_kwargs is None:
        backend_kwargs = {}

    backend_kwargs = {
        **backend_kwarg_defaults(),
        **backend_kwargs,
    }

    (figsize, _, _, _, _, _) = _scale_fig_size(figsize, None)
    backend_kwargs.setdefault("figsize", figsize)
    backend_kwargs["squeeze"] = True

    if ax is None:
        _, ax = create_axes_grid(1, backend_kwargs=backend_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}

    plot_kwargs.setdefault("where", "post")

    if fill_band:
        if fill_kwargs is None:
            fill_kwargs = {}
        fill_kwargs.setdefault("step", "post")
        fill_kwargs.setdefault("color", to_hex("C0"))
        fill_kwargs.setdefault("alpha", 0.2)
    else:
        if plot_outline_kwargs is None:
            plot_outline_kwargs = {}
        plot_outline_kwargs.setdefault("where", "post")
        plot_outline_kwargs.setdefault("color", to_hex("C0"))
        plot_outline_kwargs.setdefault("alpha", 0.2)

    ax.step(x_coord, y_coord, **plot_kwargs)

    if x_bands is not None:
        if fill_band:
            ax.fill_between(x_bands, lower, higher, **fill_kwargs)
        else:
            ax.plot(x_bands, lower, x_bands, higher, **plot_outline_kwargs)

    if backend_show(show):
        plt.show()

    return ax



"""Pareto tail indices plot."""

import logging
import warnings

import numpy as np
from xarray import DataArray

from ..rcparams import rcParams
from ..stats import ELPDData
from ..utils import get_coords
from .plot_utils import format_coords_as_labels, get_plotting_function

_log = logging.getLogger(__name__)


def plot_khat(
    khats,
    color="C0",
    xlabels=False,
    show_hlines=False,
    show_bins=False,
    bin_format="{1:.1f}%",
    annotate=False,
    threshold=None,
    hover_label=False,
    hover_format="{1}",
    figsize=None,
    textsize=None,
    coords=None,
    legend=False,
    markersize=None,
    ax=None,
    hlines_kwargs=None,
    backend=None,
    backend_kwargs=None,
    show=None,
    **kwargs
):
    r"""Plot Pareto tail indices :math:`\hat{k}` for diagnosing convergence in PSIS-LOO.

    Parameters
    ----------
    khats : ELPDData
        The input Pareto tail indices to be plotted.
    color : str or array_like, default "C0"
        Colors of the scatter plot, if color is a str all dots will have the same color,
        if it is the size of the observations, each dot will have the specified color,
        otherwise, it will be interpreted as a list of the dims to be used for the color
        code. If Matplotlib c argument is passed, it will override the color argument.
    xlabels : bool, default False
        Use coords as xticklabels.
    show_hlines : bool, default False
        Show the horizontal lines, by default at the values [0, 0.5, 0.7, 1].
    show_bins : bool, default False
        Show the percentage of khats falling in each bin, as delimited by hlines.
    bin_format : str, optional
        The string is used as formatting guide calling ``bin_format.format(count, pct)``.
    threshold : float, optional
        Show the labels of k values larger than `threshold`. If ``None`` (default), no
        observations will be highlighted.
    hover_label : bool, default False
        Show the datapoint label when hovering over it with the mouse. Requires an interactive
        backend.
    hover_format : str, default "{1}"
        String used to format the hover label via ``hover_format.format(idx, coord_label)``
    figsize : (float, float), optional
        Figure size. If ``None`` it will be defined automatically.
    textsize : float, optional
        Text size scaling factor for labels, titles and lines. If ``None`` it will be autoscaled
        based on `figsize`.
    coords : mapping, optional
        Coordinates of points to plot. **All** values are used for computation, but only a
        a subset can be plotted for convenience. See :ref:`this section <common_coords>` for
        usage examples.
    legend : bool, default False
        Include a legend to the plot. Only taken into account when color argument is a dim name.
    markersize : int, optional
        markersize for scatter plot. Defaults to ``None`` in which case it will
        be chosen based on autoscaling for figsize.
    ax : axes, optional
        Matplotlib axes or bokeh figures.
    hlines_kwargs : dict, optional
        Additional keywords passed to
        :meth:`matplotlib.axes.Axes.hlines`.
    backend : {"matplotlib", "bokeh"}, default "matplotlib"
        Select plotting backend.
    backend_kwargs : dict, optional
        These are kwargs specific to the backend being used, passed to
        :func:`matplotlib.pyplot.subplots` or :class:`bokeh.plotting.figure`.
        For additional documentation check the plotting method of the backend.
    show : bool, optional
        Call backend show function.
    kwargs :
        Additional keywords passed to
        :meth:`matplotlib.axes.Axes.scatter`.

    Returns
    -------
    axes : matplotlib_axes or bokeh_figures

    See Also
    --------
    psislw : Pareto smoothed importance sampling (PSIS).

    Examples
    --------
    Plot estimated pareto shape parameters showing how many fall in each category.

    .. plot::
        :context: close-figs

        >>> import arviz as az
        >>> radon = az.load_arviz_data("radon")
        >>> loo_radon = az.loo(radon, pointwise=True)
        >>> az.plot_khat(loo_radon, show_bins=True)

    Show xlabels

    .. plot::
        :context: close-figs

        >>> centered_eight = az.load_arviz_data("centered_eight")
        >>> khats = az.loo(centered_eight, pointwise=True).pareto_k
        >>> az.plot_khat(khats, xlabels=True, threshold=1)

    Use custom color scheme

    .. plot::
        :context: close-figs

        >>> counties = radon.posterior.County[radon.constant_data.county_idx].values
        >>> colors = [
        ...     "blue" if county[-1] in ("A", "N") else "green" for county in counties
        ... ]
        >>> az.plot_khat(loo_radon, color=colors)

    Notes
    -----
    The Generalized Pareto distribution (GPD) diagnoses convergence rates for importance
    sampling. GPD has parameters offset, scale, and shape. The shape parameter (:math:`k`)
    tells the distribution's number of finite moments. The pre-asymptotic convergence rate
    of importance sampling can be estimated based on the fractional number of finite moments
    of the importance ratio distribution. GPD is fitted to the largest importance ratios and
    interprets the estimated shape parameter :math:`k`, i.e., :math:`\hat{k}` can then be
    used as a diagnostic (most importantly if :math:`\hat{k} > 0.7`, then the convergence
    rate is impractically low). See [1]_.

    References
    ----------
    .. [1] Vehtari, A., Simpson, D., Gelman, A., Yao, Y., Gabry, J. (2024).
        Pareto Smoothed Importance Sampling. Journal of Machine Learning
        Research, 25(72):1-58.

    """
    if annotate:
        _log.warning("annotate will be deprecated, please use threshold instead")
        threshold = annotate

    if coords is None:
        coords = {}

    if color is None:
        color = "C0"

    if isinstance(khats, np.ndarray):
        warnings.warn(
            "support for arrays will be deprecated, please use ELPDData."
            "The reason for this, is that we need to know the numbers of draws"
            "sampled from the posterior",
            FutureWarning,
        )
        khats = khats.flatten()
        xlabels = False
        legend = False
        dims = []
        good_k = None
    else:
        if isinstance(khats, ELPDData):
            good_k = khats.good_k
            khats = khats.pareto_k
        else:
            good_k = None
            warnings.warn(
                "support for DataArrays will be deprecated, please use ELPDData."
                "The reason for this, is that we need to know the numbers of draws"
                "sampled from the posterior",
                FutureWarning,
            )
        if not isinstance(khats, DataArray):
            raise ValueError("Incorrect khat data input. Check the documentation")

        khats = get_coords(khats, coords)
        dims = khats.dims

    n_data_points = khats.size
    xdata = np.arange(n_data_points)
    if isinstance(khats, DataArray):
        coord_labels = format_coords_as_labels(khats)
    else:
        coord_labels = xdata.astype(str)

    plot_khat_kwargs = dict(
        hover_label=hover_label,
        hover_format=hover_format,
        ax=ax,
        figsize=figsize,
        xdata=xdata,
        khats=khats,
        good_k=good_k,
        kwargs=kwargs,
        threshold=threshold,
        coord_labels=coord_labels,
        show_hlines=show_hlines,
        show_bins=show_bins,
        hlines_kwargs=hlines_kwargs,
        xlabels=xlabels,
        legend=legend,
        color=color,
        dims=dims,
        textsize=textsize,
        markersize=markersize,
        n_data_points=n_data_points,
        bin_format=bin_format,
        backend_kwargs=backend_kwargs,
        show=show,
    )

    if backend is None:
        backend = rcParams["plot.backend"]
    backend = backend.lower()

    # TODO: Add backend kwargs
    plot = get_plotting_function("plot_khat", "khatplot", backend)
    axes = plot(**plot_khat_kwargs)
    return axes



"""Posterior/Prior predictive plot."""

import logging
import warnings
from numbers import Integral

import numpy as np

from ..labels import BaseLabeller
from ..sel_utils import xarray_var_iter
from ..rcparams import rcParams
from ..utils import _var_names
from .plot_utils import default_grid, filter_plotters_list, get_plotting_function

_log = logging.getLogger(__name__)


def plot_ppc(
    data,
    kind="kde",
    alpha=None,
    mean=True,
    observed=None,
    observed_rug=False,
    color=None,
    colors=None,
    grid=None,
    figsize=None,
    textsize=None,
    data_pairs=None,
    var_names=None,
    filter_vars=None,
    coords=None,
    flatten=None,
    flatten_pp=None,
    num_pp_samples=None,
    random_seed=None,
    jitter=None,
    animated=False,
    animation_kwargs=None,
    legend=True,
    labeller=None,
    ax=None,
    backend=None,
    backend_kwargs=None,
    group="posterior",
    show=None,
):
    """
    Plot for posterior/prior predictive checks.

    Parameters
    ----------
    data : InferenceData
        :class:`arviz.InferenceData` object containing the observed and posterior/prior
        predictive data.
    kind : str, default "kde"
        Type of plot to display ("kde", "cumulative", or "scatter").
    alpha : float, optional
        Opacity of posterior/prior predictive density curves.
        Defaults to 0.2 for ``kind = kde`` and cumulative, for scatter defaults to 0.7.
    mean : bool, default True
        Whether or not to plot the mean posterior/prior predictive distribution.
    observed : bool, optional
        Whether or not to plot the observed data. Defaults to True for ``group = posterior``
        and False for ``group = prior``.
    observed_rug : bool, default False
        Whether or not to plot a rug plot for the observed data. Only valid if `observed` is
        `True` and for kind `kde` or `cumulative`.
    color : list, optional
        List with valid matplotlib colors corresponding to the posterior/prior predictive
        distribution, observed data and mean of the posterior/prior predictive distribution.
        Defaults to ["C0", "k", "C1"].
    grid : tuple, optional
        Number of rows and columns. Defaults to None, the rows and columns are
        automatically inferred.
    figsize : tuple, optional
        Figure size. If None, it will be defined automatically.
    textsize : float, optional
        Text size scaling factor for labels, titles and lines. If None, it will be
        autoscaled based on ``figsize``.
    data_pairs : dict, optional
        Dictionary containing relations between observed data and posterior/prior predictive data.
        Dictionary structure:

        - key = data var_name
        - value = posterior/prior predictive var_name

        For example, ``data_pairs = {'y' : 'y_hat'}``
        If None, it will assume that the observed data and the posterior/prior
        predictive data have the same variable name.
    var_names : list of str, optional
        Variables to be plotted, if `None` all variable are plotted. Prefix the
        variables by ``~`` when you want to exclude them from the plot.
    filter_vars : {None, "like", "regex"}, default None
        If `None` (default), interpret var_names as the real variables names. If "like",
        interpret var_names as substrings of the real variables names. If "regex",
        interpret var_names as regular expressions on the real variables names. A la
        ``pandas.filter``.
    coords : dict, optional
        Dictionary mapping dimensions to selected coordinates to be plotted.
        Dimensions without a mapping specified will include all coordinates for
        that dimension. Defaults to including all coordinates for all
        dimensions if None.
    flatten : list
        List of dimensions to flatten in ``observed_data``. Only flattens across the coordinates
        specified in the ``coords`` argument. Defaults to flattening all of the dimensions.
    flatten_pp : list
        List of dimensions to flatten in posterior_predictive/prior_predictive. Only flattens
        across the coordinates specified in the ``coords`` argument. Defaults to flattening all
        of the dimensions. Dimensions should match flatten excluding dimensions for ``data_pairs``
        parameters. If ``flatten`` is defined and ``flatten_pp`` is None, then
        ``flatten_pp = flatten``.
    num_pp_samples : int
        The number of posterior/prior predictive samples to plot. For ``kind`` = 'scatter' and
        ``animation = False`` if defaults to a maximum of 5 samples and will set jitter to 0.7.
        unless defined. Otherwise it defaults to all provided samples.
    random_seed : int
        Random number generator seed passed to ``numpy.random.seed`` to allow
        reproducibility of the plot. By default, no seed will be provided
        and the plot will change each call if a random sample is specified
        by ``num_pp_samples``.
    jitter : float, default 0
        If ``kind`` is "scatter", jitter will add random uniform noise to the height
        of the ppc samples and observed data.
    animated : bool, default False
        Create an animation of one posterior/prior predictive sample per frame.
        Only works with matploblib backend.
        To run animations inside a notebook you have to use the `nbAgg` matplotlib's backend.
        Try with `%matplotlib notebook` or  `%matplotlib  nbAgg`. You can switch back to the
        default matplotlib's backend with `%matplotlib  inline` or `%matplotlib  auto`.
        If switching back and forth between matplotlib's backend, you may need to run twice the cell
        with the animation.
        If you experience problems rendering the animation try setting
        ``animation_kwargs({'blit':False})`` or changing the matplotlib's backend (e.g. to TkAgg)
        If you run the animation from a script write ``ax, ani = az.plot_ppc(.)``
    animation_kwargs : dict
        Keywords passed to  :class:`matplotlib.animation.FuncAnimation`. Ignored with
        matplotlib backend.
    legend : bool, default True
        Add legend to figure.
    labeller : labeller, optional
        Class providing the method ``make_pp_label`` to generate the labels in the plot titles.
        Read the :ref:`label_guide` for more details and usage examples.
    ax : numpy array-like of matplotlib_axes or bokeh figures, optional
        A 2D array of locations into which to plot the densities. If not supplied, Arviz will create
        its own array of plot areas (and return it).
    backend : str, optional
        Select plotting backend {"matplotlib","bokeh"}. Default to "matplotlib".
    backend_kwargs : dict, optional
        These are kwargs specific to the backend being used, passed to
        :func:`matplotlib.pyplot.subplots` or :func:`bokeh.plotting.figure`.
        For additional documentation check the plotting method of the backend.
    group : {"prior", "posterior"}, optional
        Specifies which InferenceData group should be plotted. Defaults to 'posterior'.
        Other value can be 'prior'.
    show : bool, optional
        Call backend show function.

    Returns
    -------
    axes : matplotlib_axes or bokeh_figures
    ani : matplotlib.animation.FuncAnimation, optional
        Only provided if `animated` is ``True``.

    See Also
    --------
    plot_bpv : Plot Bayesian p-value for observed data and Posterior/Prior predictive.
    plot_loo_pit : Plot for posterior predictive checks using cross validation.
    plot_lm : Posterior predictive and mean plots for regression-like data.
    plot_ts : Plot timeseries data.

    Examples
    --------
    Plot the observed data KDE overlaid on posterior predictive KDEs.

    .. plot::
        :context: close-figs

        >>> import arviz as az
        >>> data = az.load_arviz_data('radon')
        >>> az.plot_ppc(data, data_pairs={"y":"y"})

    Plot the overlay with empirical CDFs.

    .. plot::
        :context: close-figs

        >>> az.plot_ppc(data, kind='cumulative')

    Use the ``coords`` and ``flatten`` parameters to plot selected variable dimensions
    across multiple plots. We will now modify the dimension ``obs_id`` to contain
    indicate the name of the county where the measure was taken. The change has to
    be done on both ``posterior_predictive`` and ``observed_data`` groups, which is
    why we will use :meth:`~arviz.InferenceData.map` to apply the same function to
    both groups. Afterwards, we will select the counties to be plotted with the
    ``coords`` arg.

    .. plot::
        :context: close-figs

        >>> obs_county = data.posterior["County"][data.constant_data["county_idx"]]
        >>> data = data.assign_coords(obs_id=obs_county, groups="observed_vars")
        >>> az.plot_ppc(data, coords={'obs_id': ['ANOKA', 'BELTRAMI']}, flatten=[])

    Plot the overlay using a stacked scatter plot that is particularly useful
    when the sample sizes are small.

    .. plot::
        :context: close-figs

        >>> az.plot_ppc(data, kind='scatter', flatten=[],
        >>>             coords={'obs_id': ['AITKIN', 'BELTRAMI']})

    Plot random posterior predictive sub-samples.

    .. plot::
        :context: close-figs

        >>> az.plot_ppc(data, num_pp_samples=30, random_seed=7)
    """
    if group not in ("posterior", "prior"):
        raise TypeError("`group` argument must be either `posterior` or `prior`")

    for groups in (f"{group}_predictive", "observed_data"):
        if not hasattr(data, groups):
            raise TypeError(f'`data` argument must have the group "{groups}" for ppcplot')

    if kind.lower() not in ("kde", "cumulative", "scatter"):
        raise TypeError("`kind` argument must be either `kde`, `cumulative`, or `scatter`")

    if colors is None:
        colors = ["C0", "k", "C1"]

    if isinstance(colors, str):
        raise TypeError("colors should be a list with 3 items.")

    if len(colors) != 3:
        raise ValueError("colors should be a list with 3 items.")

    if color is not None:
        warnings.warn("color has been deprecated in favor of colors", FutureWarning)
        colors[0] = color

    if data_pairs is None:
        data_pairs = {}

    if backend is None:
        backend = rcParams["plot.backend"]
    backend = backend.lower()
    if backend == "bokeh" and animated:
        raise TypeError("Animation option is only supported with matplotlib backend.")

    observed_data = data.observed_data

    if group == "posterior":
        predictive_dataset = data.posterior_predictive
        if observed is None:
            observed = True
    elif group == "prior":
        predictive_dataset = data.prior_predictive
        if observed is None:
            observed = False

    if var_names is None:
        var_names = list(observed_data.data_vars)
    var_names = _var_names(var_names, observed_data, filter_vars)
    pp_var_names = [data_pairs.get(var, var) for var in var_names]
    pp_var_names = _var_names(pp_var_names, predictive_dataset, filter_vars)

    if flatten_pp is None:
        if flatten is None:
            flatten_pp = list(predictive_dataset.dims)
        else:
            flatten_pp = flatten
    if flatten is None:
        flatten = list(observed_data.dims)

    if coords is None:
        coords = {}
    else:
        coords = coords.copy()

    if labeller is None:
        labeller = BaseLabeller()

    if random_seed is not None:
        np.random.seed(random_seed)

    total_pp_samples = predictive_dataset.sizes["chain"] * predictive_dataset.sizes["draw"]
    if num_pp_samples is None:
        if kind == "scatter" and not animated:
            num_pp_samples = min(5, total_pp_samples)
        else:
            num_pp_samples = total_pp_samples

    if (
        not isinstance(num_pp_samples, Integral)
        or num_pp_samples < 1
        or num_pp_samples > total_pp_samples
    ):
        raise TypeError(f"`num_pp_samples` must be an integer between 1 and {total_pp_samples}.")

    pp_sample_ix = np.random.choice(total_pp_samples, size=num_pp_samples, replace=False)

    for key in coords.keys():
        coords[key] = np.where(np.isin(observed_data[key], coords[key]))[0]

    obs_plotters = filter_plotters_list(
        list(
            xarray_var_iter(
                observed_data.isel(coords),
                skip_dims=set(flatten),
                var_names=var_names,
                combined=True,
                dim_order=["chain", "draw"],
            )
        ),
        "plot_ppc",
    )
    length_plotters = len(obs_plotters)
    pp_plotters = [
        tup
        for _, tup in zip(
            range(length_plotters),
            xarray_var_iter(
                predictive_dataset.isel(coords),
                var_names=pp_var_names,
                skip_dims=set(flatten_pp),
                combined=True,
                dim_order=["chain", "draw"],
            ),
        )
    ]
    rows, cols = default_grid(length_plotters, grid=grid)

    ppcplot_kwargs = dict(
        ax=ax,
        length_plotters=length_plotters,
        rows=rows,
        cols=cols,
        figsize=figsize,
        animated=animated,
        obs_plotters=obs_plotters,
        pp_plotters=pp_plotters,
        predictive_dataset=predictive_dataset,
        pp_sample_ix=pp_sample_ix,
        kind=kind,
        alpha=alpha,
        colors=colors,
        jitter=jitter,
        textsize=textsize,
        mean=mean,
        observed=observed,
        observed_rug=observed_rug,
        total_pp_samples=total_pp_samples,
        legend=legend,
        labeller=labeller,
        group=group,
        animation_kwargs=animation_kwargs,
        num_pp_samples=num_pp_samples,
        backend_kwargs=backend_kwargs,
        show=show,
    )

    # TODO: Add backend kwargs
    plot = get_plotting_function("plot_ppc", "ppcplot", backend)
    axes = plot(**ppcplot_kwargs)
    return axes



"""Plot ecdf or ecdf-difference plot with confidence bands."""

import warnings

import numpy as np
from scipy.stats import uniform

try:
    from scipy.stats import ecdf as scipy_ecdf
except ImportError:
    scipy_ecdf = None

from ..rcparams import rcParams
from ..stats.ecdf_utils import ecdf_confidence_band, _get_ecdf_points
from ..utils import BehaviourChangeWarning
from .plot_utils import get_plotting_function


def plot_ecdf(
    values,
    values2=None,
    eval_points=None,
    cdf=None,
    difference=False,
    confidence_bands=False,
    ci_prob=None,
    num_trials=500,
    rvs=None,
    random_state=None,
    figsize=None,
    fill_band=True,
    plot_kwargs=None,
    fill_kwargs=None,
    plot_outline_kwargs=None,
    ax=None,
    show=None,
    backend=None,
    backend_kwargs=None,
    npoints=100,
    pointwise=False,
    fpr=None,
    pit=False,
    **kwargs,
):
    r"""Plot ECDF or ECDF-Difference Plot with Confidence bands.

    Plots of the empirical cumulative distribution function (ECDF) of an array. Optionally, A `cdf`
    argument representing a reference CDF may be provided for comparison using a difference ECDF
    plot and/or confidence bands.

    Alternatively, the PIT for a single dataset may be visualized.

    Notes
    -----
    This plot computes the confidence bands with the simulated based algorithm presented in [1]_.

    Parameters
    ----------
    values : array-like
        Values to plot from an unknown continuous or discrete distribution.
    values2 : array-like, optional
        values to compare to the original sample.

        .. deprecated:: 0.18.0
           Instead use ``cdf=scipy.stats.ecdf(values2).cdf.evaluate``.
    cdf : callable, optional
        Cumulative distribution function of the distribution to compare the original sample.
        The function must take as input a numpy array of draws from the distribution.
    difference : bool, default False
        If True then plot ECDF-difference plot otherwise ECDF plot.
    confidence_bands : str or bool

        - False: No confidence bands are plotted (default).
        - True: Plot bands computed with the default algorithm (subject to change)
        - "pointwise": Compute the pointwise (i.e. marginal) confidence band.
        - "optimized": Use optimization to estimate a simultaneous confidence band.
        - "simulated": Use Monte Carlo simulation to estimate a simultaneous confidence
          band.

        For simultaneous confidence bands to be correctly calibrated, provide `eval_points` that
        are not dependent on the `values`.
    ci_prob : float, default 0.94
        The probability that the true ECDF lies within the confidence band. If `confidence_bands`
        is "pointwise", this is the marginal probability instead of the joint probability.
    eval_points : array-like, optional
        The points at which to evaluate the ECDF. If None, `npoints` uniformly spaced points
        between the data bounds will be used.
    rvs: callable, optional
        A function that takes an integer `ndraws` and optionally the object passed to
        `random_state` and returns an array of `ndraws` samples from the same distribution
        as the original dataset. Required if `method` is "simulated" and variable is discrete.
    random_state : int, numpy.random.Generator or numpy.random.RandomState, optional
    num_trials : int, default 500
        The number of random ECDFs to generate for constructing simultaneous confidence bands
        (if `confidence_bands` is "simulated").
    figsize : (float,float), optional
        Figure size. If `None` it will be defined automatically.
    fill_band : bool, default True
        If True it fills in between to mark the area inside the confidence interval. Otherwise,
        plot the border lines.
    plot_kwargs : dict, optional
        Additional kwargs passed to :func:`mpl:matplotlib.pyplot.step` or
        :meth:`bokeh.plotting.figure.step`
    fill_kwargs : dict, optional
        Additional kwargs passed to :func:`mpl:matplotlib.pyplot.fill_between` or
        :meth:`bokeh:bokeh.plotting.Figure.varea`
    plot_outline_kwargs : dict, optional
        Additional kwargs passed to :meth:`mpl:matplotlib.axes.Axes.plot` or
        :meth:`bokeh:bokeh.plotting.Figure.line`
    ax :axes, optional
        Matplotlib axes or bokeh figures.
    show : bool, optional
        Call backend show function.
    backend : {"matplotlib", "bokeh"}, default "matplotlib"
        Select plotting backend.
    backend_kwargs : dict, optional
        These are kwargs specific to the backend being used, passed to
        :func:`matplotlib.pyplot.subplots` or :class:`bokeh.plotting.figure`.
        For additional documentation check the plotting method of the backend.
    npoints : int, default 100
        The number of evaluation points for the ecdf or ecdf-difference plots, if `eval_points` is
        not provided or `pit` is `True`.

        .. deprecated:: 0.18.0
           Instead specify ``eval_points=np.linspace(np.min(values), np.max(values), npoints)``
           unless `pit` is `True`.
    pointwise : bool, default False

        .. deprecated:: 0.18.0
           Instead use `confidence_bands="pointwise"`.
    fpr : float, optional

        .. deprecated:: 0.18.0
           Instead use `ci_prob=1-fpr`.
    pit : bool, default False
        If True plots the ECDF or ECDF-diff of PIT of sample.

        .. deprecated:: 0.18.0
           See below example instead.

    Returns
    -------
    axes : matplotlib_axes or bokeh_figure

    References
    ----------
    .. [1] Säilynoja, T., Bürkner, P.C. and Vehtari, A. (2022). Graphical Test for
        Discrete Uniformity and its Applications in Goodness of Fit Evaluation and
        Multiple Sample Comparison. Statistics and Computing, 32(32).

    Examples
    --------
    In a future release, the default behaviour of ``plot_ecdf`` will change.
    To maintain the original behaviour you should do:

    .. plot::
        :context: close-figs

        >>> import arviz as az
        >>> import numpy as np
        >>> from scipy.stats import uniform, norm
        >>>
        >>> sample = norm(0,1).rvs(1000)
        >>> npoints = 100
        >>> az.plot_ecdf(sample, eval_points=np.linspace(sample.min(), sample.max(), npoints))

    However, seeing this warning isn't an indicator of anything being wrong,
    if you are happy to get different behaviour as ArviZ improves and adds
    new algorithms you can ignore it like so:

    .. plot::
        :context: close-figs

        >>> import warnings
        >>> warnings.filterwarnings("ignore", category=az.utils.BehaviourChangeWarning)

    Plot an ECDF plot for a given sample evaluated at the sample points. This will become
    the new behaviour when `eval_points` is not provided:

    .. plot::
        :context: close-figs

        >>> az.plot_ecdf(sample, eval_points=np.unique(sample))

    Plot an ECDF plot with confidence bands for comparing a given sample to a given distribution.
    We manually specify evaluation points independent of the values so that the confidence bands
    are correctly calibrated.

    .. plot::
        :context: close-figs

        >>> distribution = norm(0,1)
        >>> eval_points = np.linspace(*distribution.ppf([0.001, 0.999]), 100)
        >>> az.plot_ecdf(
        >>>     sample, eval_points=eval_points,
        >>>     cdf=distribution.cdf, confidence_bands=True
        >>> )

    Plot an ECDF-difference plot with confidence bands for comparing a given sample
    to a given distribution.

    .. plot::
        :context: close-figs

        >>> az.plot_ecdf(
        >>>     sample, cdf=distribution.cdf,
        >>>     confidence_bands=True, difference=True
        >>> )

    Plot an ECDF plot with confidence bands for the probability integral transform (PIT) of a
    continuous sample. If drawn from the reference distribution, the PIT values should be uniformly
    distributed.

    .. plot::
        :context: close-figs

        >>> pit_vals = distribution.cdf(sample)
        >>> uniform_dist = uniform(0, 1)
        >>> az.plot_ecdf(
        >>>     pit_vals, cdf=uniform_dist.cdf, confidence_bands=True,
        >>> )

    Plot an ECDF-difference plot of PIT values.

    .. plot::
        :context: close-figs

        >>> az.plot_ecdf(
        >>>     pit_vals, cdf = uniform_dist.cdf, confidence_bands = True,
        >>>     difference = True
        >>> )
    """
    if confidence_bands is True:
        if pointwise:
            warnings.warn(
                "`pointwise` has been deprecated. Use `confidence_bands='pointwise'` instead.",
                FutureWarning,
            )
            confidence_bands = "pointwise"
        else:
            confidence_bands = "auto"
        # if pointwise specified, confidence_bands must be a bool or 'pointwise'
    elif confidence_bands not in [False, "pointwise"] and pointwise:
        raise ValueError(
            f"Cannot specify both `confidence_bands='{confidence_bands}'` and `pointwise=True`"
        )

    if fpr is not None:
        warnings.warn(
            "`fpr` has been deprecated. Use `ci_prob=1-fpr` or set `rcParam['stats.ci_prob']` to"
            "`1-fpr`.",
            FutureWarning,
        )
        if ci_prob is not None:
            raise ValueError("Cannot specify both `fpr` and `ci_prob`")
        ci_prob = 1 - fpr

    if ci_prob is None:
        ci_prob = rcParams["stats.ci_prob"]

    if values2 is not None:
        if cdf is not None:
            raise ValueError("You cannot specify both `values2` and `cdf`")
        if scipy_ecdf is None:
            raise ValueError(
                "The `values2` argument is deprecated and `scipy.stats.ecdf` is not available. "
                "Please use `cdf` instead."
            )
        warnings.warn(
            "`values2` has been deprecated. Use `cdf=scipy.stats.ecdf(values2).cdf.evaluate` "
            "instead.",
            FutureWarning,
        )
        cdf = scipy_ecdf(np.ravel(values2)).cdf.evaluate

    if cdf is None:
        if confidence_bands:
            raise ValueError("For confidence bands you must specify cdf")
        if difference is True:
            raise ValueError("For ECDF difference plot you must specify cdf")
        if pit:
            raise ValueError("For PIT plot you must specify cdf")

    values = np.ravel(values)
    values.sort()

    if pit:
        warnings.warn(
            "`pit` has been deprecated. Specify `values=cdf(values)` instead.",
            FutureWarning,
        )
        values = cdf(values)
        cdf = uniform(0, 1).cdf
        rvs = uniform(0, 1).rvs
        eval_points = np.linspace(1 / npoints, 1, npoints)

    if eval_points is None:
        warnings.warn(
            "In future versions, if `eval_points` is not provided, then the ECDF will be evaluated"
            " at the unique values of the sample. To keep the current behavior, provide "
            "`eval_points` explicitly.",
            BehaviourChangeWarning,
        )
        if confidence_bands in ["optimized", "simulated"]:
            warnings.warn(
                "For simultaneous bands to be correctly calibrated, specify `eval_points` "
                "independent of the `values`"
            )
        eval_points = np.linspace(values[0], values[-1], npoints)
    else:
        eval_points = np.asarray(eval_points)

    if difference or confidence_bands:
        cdf_at_eval_points = cdf(eval_points)
    else:
        cdf_at_eval_points = np.zeros_like(eval_points)

    x_coord, y_coord = _get_ecdf_points(values, eval_points, difference)

    if difference:
        y_coord -= cdf_at_eval_points

    if confidence_bands:
        ndraws = len(values)
        if confidence_bands == "auto":
            if ndraws < 200 or num_trials >= 250 * np.sqrt(ndraws):
                confidence_bands = "optimized"
            else:
                confidence_bands = "simulated"
        x_bands = eval_points
        lower, higher = ecdf_confidence_band(
            ndraws,
            eval_points,
            cdf_at_eval_points,
            method=confidence_bands,
            prob=ci_prob,
            num_trials=num_trials,
            rvs=rvs,
            random_state=random_state,
        )

        if difference:
            lower -= cdf_at_eval_points
            higher -= cdf_at_eval_points
    else:
        x_bands, lower, higher = None, None, None

    ecdf_plot_args = dict(
        x_coord=x_coord,
        y_coord=y_coord,
        x_bands=x_bands,
        lower=lower,
        higher=higher,
        figsize=figsize,
        fill_band=fill_band,
        plot_kwargs=plot_kwargs,
        fill_kwargs=fill_kwargs,
        plot_outline_kwargs=plot_outline_kwargs,
        ax=ax,
        show=show,
        backend_kwargs=backend_kwargs,
        **kwargs,
    )

    if backend is None:
        backend = rcParams["plot.backend"]
    backend = backend.lower()

    plot = get_plotting_function("plot_ecdf", "ecdfplot", backend)
    ax = plot(**ecdf_plot_args)

    return ax



"""Bokeh dotplot."""

import math
import warnings
import numpy as np

from ...plot_utils import _scale_fig_size, vectorized_to_hex
from .. import show_layout
from . import create_axes_grid
from ...plot_utils import plot_point_interval
from ...dotplot import wilkinson_algorithm, layout_stacks


def plot_dot(
    values,
    binwidth,
    dotsize,
    stackratio,
    hdi_prob,
    quartiles,
    rotated,
    dotcolor,
    intervalcolor,
    markersize,
    markercolor,
    marker,
    figsize,
    linewidth,
    point_estimate,
    nquantiles,
    point_interval,
    ax,
    show,
    backend_kwargs,
    plot_kwargs,
):
    """Bokeh dotplot."""
    if backend_kwargs is None:
        backend_kwargs = {}

    backend_kwargs.setdefault("match_aspect", True)

    (figsize, _, _, _, auto_linewidth, auto_markersize) = _scale_fig_size(figsize, None)
    dotcolor = vectorized_to_hex(dotcolor)
    intervalcolor = vectorized_to_hex(intervalcolor)
    markercolor = vectorized_to_hex(markercolor)

    if plot_kwargs is None:
        plot_kwargs = {}
    else:
        plot_kwargs = plot_kwargs.copy()
    plot_kwargs.setdefault("color", dotcolor)
    plot_kwargs.setdefault("marker", "circle")

    if linewidth is None:
        linewidth = auto_linewidth

    if markersize is None:
        markersize = auto_markersize

    if ax is None:
        ax = create_axes_grid(
            1,
            figsize=figsize,
            squeeze=True,
            backend_kwargs=backend_kwargs,
        )

    if point_interval:
        ax = plot_point_interval(
            ax,
            values,
            point_estimate,
            hdi_prob,
            quartiles,
            linewidth,
            markersize,
            markercolor,
            marker,
            rotated,
            intervalcolor,
            "bokeh",
        )

    if nquantiles > values.shape[0]:
        warnings.warn(
            "nquantiles must be less than or equal to the number of data points", UserWarning
        )
        nquantiles = values.shape[0]
    else:
        qlist = np.linspace(1 / (2 * nquantiles), 1 - 1 / (2 * nquantiles), nquantiles)
        values = np.quantile(values, qlist)

    if binwidth is None:
        binwidth = math.sqrt((values[-1] - values[0] + 1) ** 2 / (2 * nquantiles * np.pi))

    ## Wilkinson's Algorithm
    stack_locs, stack_count = wilkinson_algorithm(values, binwidth)
    x, y = layout_stacks(stack_locs, stack_count, binwidth, stackratio, rotated)

    ax.scatter(x, y, radius=dotsize * (binwidth / 2), **plot_kwargs, radius_dimension="y")
    if rotated:
        ax.xaxis.major_tick_line_color = None
        ax.xaxis.minor_tick_line_color = None
        ax.xaxis.major_label_text_font_size = "0pt"
    else:
        ax.yaxis.major_tick_line_color = None
        ax.yaxis.minor_tick_line_color = None
        ax.yaxis.major_label_text_font_size = "0pt"

    show_layout(ax, show)

    return ax



"""Bokeh ecdfplot."""

from matplotlib.colors import to_hex

from ...plot_utils import _scale_fig_size
from .. import show_layout
from . import backend_kwarg_defaults, create_axes_grid


def plot_ecdf(
    x_coord,
    y_coord,
    x_bands,
    lower,
    higher,
    plot_kwargs,
    fill_kwargs,
    plot_outline_kwargs,
    figsize,
    fill_band,
    ax,
    show,
    backend_kwargs,
):
    """Bokeh ecdfplot."""
    if backend_kwargs is None:
        backend_kwargs = {}

    backend_kwargs = {
        **backend_kwarg_defaults(),
        **backend_kwargs,
    }

    (figsize, *_) = _scale_fig_size(figsize, None)

    if ax is None:
        ax = create_axes_grid(
            1,
            figsize=figsize,
            squeeze=True,
            backend_kwargs=backend_kwargs,
        )

    if plot_kwargs is None:
        plot_kwargs = {}

    plot_kwargs.setdefault("mode", "after")

    if fill_band:
        if fill_kwargs is None:
            fill_kwargs = {}
        fill_kwargs.setdefault("fill_color", to_hex("C0"))
        fill_kwargs.setdefault("fill_alpha", 0.2)
    else:
        if plot_outline_kwargs is None:
            plot_outline_kwargs = {}
        plot_outline_kwargs.setdefault("color", to_hex("C0"))
        plot_outline_kwargs.setdefault("alpha", 0.2)

    if x_bands is not None:
        ax.step(x_coord, y_coord, **plot_kwargs)

        if fill_band:
            ax.varea(x_bands, lower, higher, **fill_kwargs)
        else:
            ax.line(x_bands, lower, **plot_outline_kwargs)
            ax.line(x_bands, higher, **plot_outline_kwargs)
    else:
        ax.step(x_coord, y_coord, **plot_kwargs)

    show_layout(ax, show)

    return ax



# pylint: disable=no-member, invalid-name, redefined-outer-name, too-many-function-args
import importlib
from collections import OrderedDict
import os

import numpy as np
import pytest

from ... import from_pystan

from ...data.io_pystan import get_draws, get_draws_stan3  # pylint: disable=unused-import
from ..helpers import (  # pylint: disable=unused-import
    chains,
    check_multiple_attrs,
    draws,
    eight_schools_params,
    importorskip,
    load_cached_models,
    pystan_version,
)

# Check if either pystan or pystan3 is installed
pystan_installed = (importlib.util.find_spec("pystan") is not None) or (
    importlib.util.find_spec("stan") is not None
)


@pytest.mark.skipif(
    not (pystan_installed or "ARVIZ_REQUIRE_ALL_DEPS" in os.environ),
    reason="test requires pystan/pystan3 which is not installed",
)
class TestDataPyStan:
    @pytest.fixture(scope="class")
    def data(self, eight_schools_params, draws, chains):
        class Data:
            model, obj = load_cached_models(eight_schools_params, draws, chains, "pystan")["pystan"]

        return Data

    def get_inference_data(self, data, eight_schools_params):
        """vars as str."""
        return from_pystan(
            posterior=data.obj,
            posterior_predictive="y_hat",
            predictions="y_hat",  # wrong, but fine for testing
            prior=data.obj,
            prior_predictive="y_hat",
            observed_data="y",
            constant_data="sigma",
            predictions_constant_data="sigma",  # wrong, but fine for testing
            log_likelihood={"y": "log_lik"},
            coords={"school": np.arange(eight_schools_params["J"])},
            dims={
                "theta": ["school"],
                "y": ["school"],
                "sigma": ["school"],
                "y_hat": ["school"],
                "eta": ["school"],
            },
            posterior_model=data.model,
            prior_model=data.model,
        )

    def get_inference_data2(self, data, eight_schools_params):
        """vars as lists."""
        return from_pystan(
            posterior=data.obj,
            posterior_predictive=["y_hat"],
            predictions=["y_hat"],  # wrong, but fine for testing
            prior=data.obj,
            prior_predictive=["y_hat"],
            observed_data=["y"],
            log_likelihood="log_lik",
            coords={
                "school": np.arange(eight_schools_params["J"]),
                "log_likelihood_dim": np.arange(eight_schools_params["J"]),
            },
            dims={
                "theta": ["school"],
                "y": ["school"],
                "y_hat": ["school"],
                "eta": ["school"],
                "log_lik": ["log_likelihood_dim"],
            },
            posterior_model=data.model,
            prior_model=data.model,
        )

    def get_inference_data3(self, data, eight_schools_params):
        """multiple vars as lists."""
        return from_pystan(
            posterior=data.obj,
            posterior_predictive=["y_hat", "log_lik"],  # wrong, but fine for testing
            predictions=["y_hat", "log_lik"],  # wrong, but fine for testing
            prior=data.obj,
            prior_predictive=["y_hat", "log_lik"],  # wrong, but fine for testing
            constant_data=["sigma", "y"],  # wrong, but fine for testing
            predictions_constant_data=["sigma", "y"],  # wrong, but fine for testing
            coords={"school": np.arange(eight_schools_params["J"])},
            dims={
                "theta": ["school"],
                "y": ["school"],
                "sigma": ["school"],
                "y_hat": ["school"],
                "eta": ["school"],
            },
            posterior_model=data.model,
            prior_model=data.model,
        )

    def get_inference_data4(self, data):
        """minimal input."""
        return from_pystan(
            posterior=data.obj,
            posterior_predictive=None,
            prior=data.obj,
            prior_predictive=None,
            coords=None,
            dims=None,
            posterior_model=data.model,
            log_likelihood=[],
            prior_model=data.model,
            save_warmup=True,
        )

    def get_inference_data5(self, data):
        """minimal input."""
        return from_pystan(
            posterior=data.obj,
            posterior_predictive=None,
            prior=data.obj,
            prior_predictive=None,
            coords=None,
            dims=None,
            posterior_model=data.model,
            log_likelihood=False,
            prior_model=data.model,
            save_warmup=True,
            dtypes={"eta": int},
        )

    def test_sampler_stats(self, data, eight_schools_params):
        inference_data = self.get_inference_data(data, eight_schools_params)
        test_dict = {"sample_stats": ["diverging"]}
        fails = check_multiple_attrs(test_dict, inference_data)
        assert not fails

    def test_inference_data(self, data, eight_schools_params):
        inference_data1 = self.get_inference_data(data, eight_schools_params)
        inference_data2 = self.get_inference_data2(data, eight_schools_params)
        inference_data3 = self.get_inference_data3(data, eight_schools_params)
        inference_data4 = self.get_inference_data4(data)
        inference_data5 = self.get_inference_data5(data)
        # inference_data 1
        test_dict = {
            "posterior": ["theta", "~log_lik"],
            "posterior_predictive": ["y_hat"],
            "predictions": ["y_hat"],
            "observed_data": ["y"],
            "constant_data": ["sigma"],
            "predictions_constant_data": ["sigma"],
            "sample_stats": ["diverging", "lp"],
            "log_likelihood": ["y", "~log_lik"],
            "prior": ["theta"],
        }
        fails = check_multiple_attrs(test_dict, inference_data1)
        assert not fails
        # inference_data 2
        test_dict = {
            "posterior_predictive": ["y_hat"],
            "predictions": ["y_hat"],
            "observed_data": ["y"],
            "sample_stats_prior": ["diverging"],
            "sample_stats": ["diverging", "lp"],
            "log_likelihood": ["log_lik"],
            "prior_predictive": ["y_hat"],
        }
        fails = check_multiple_attrs(test_dict, inference_data2)
        assert not fails
        assert any(
            item in inference_data2.posterior.attrs for item in ["stan_code", "program_code"]
        )
        assert any(
            item in inference_data2.sample_stats.attrs for item in ["stan_code", "program_code"]
        )
        # inference_data 3
        test_dict = {
            "posterior_predictive": ["y_hat", "log_lik"],
            "predictions": ["y_hat", "log_lik"],
            "constant_data": ["sigma", "y"],
            "predictions_constant_data": ["sigma", "y"],
            "sample_stats_prior": ["diverging"],
            "sample_stats": ["diverging", "lp"],
            "log_likelihood": ["log_lik"],
            "prior_predictive": ["y_hat", "log_lik"],
        }
        fails = check_multiple_attrs(test_dict, inference_data3)
        assert not fails
        # inference_data 4
        test_dict = {
            "posterior": ["theta"],
            "prior": ["theta"],
            "sample_stats": ["diverging", "lp"],
            "~log_likelihood": [""],
            "warmup_posterior": ["theta"],
            "warmup_sample_stats": ["diverging", "lp"],
        }
        fails = check_multiple_attrs(test_dict, inference_data4)
        assert not fails
        # inference_data 5
        test_dict = {
            "posterior": ["theta"],
            "prior": ["theta"],
            "sample_stats": ["diverging", "lp"],
            "~log_likelihood": [""],
            "warmup_posterior": ["theta"],
            "warmup_sample_stats": ["diverging", "lp"],
        }
        fails = check_multiple_attrs(test_dict, inference_data5)
        assert not fails
        assert inference_data5.posterior.eta.dtype.kind == "i"

    def test_invalid_fit(self, data):
        if pystan_version() == 2:
            model = data.model
            model_data = {
                "J": 8,
                "y": np.array([28.0, 8.0, -3.0, 7.0, -1.0, 1.0, 18.0, 12.0]),
                "sigma": np.array([15.0, 10.0, 16.0, 11.0, 9.0, 11.0, 10.0, 18.0]),
            }
            fit_test_grad = model.sampling(
                data=model_data, test_grad=True, check_hmc_diagnostics=False
            )
            with pytest.raises(AttributeError):
                _ = from_pystan(posterior=fit_test_grad)
            fit = model.sampling(data=model_data, iter=100, chains=1, check_hmc_diagnostics=False)
            del fit.sim["samples"]
            with pytest.raises(AttributeError):
                _ = from_pystan(posterior=fit)

    def test_empty_parameter(self):
        model_code = """
            parameters {
                real y;
                vector[3] x;
                vector[0] a;
                vector[2] z;
            }
            model {
                y ~ normal(0,1);
            }
        """
        if pystan_version() == 2:
            from pystan import StanModel  # pylint: disable=import-error

            model = StanModel(model_code=model_code)
            fit = model.sampling(iter=500, chains=2, check_hmc_diagnostics=False)
        else:
            import stan  # pylint: disable=import-error

            model = stan.build(model_code)
            fit = model.sample(num_samples=500, num_chains=2)

        posterior = from_pystan(posterior=fit)
        test_dict = {"posterior": ["y", "x", "z", "~a"], "sample_stats": ["diverging"]}
        fails = check_multiple_attrs(test_dict, posterior)
        assert not fails

    def test_get_draws(self, data):
        fit = data.obj
        if pystan_version() == 2:
            draws, _ = get_draws(fit, variables=["theta", "theta"])
        else:
            draws, _ = get_draws_stan3(fit, variables=["theta", "theta"])
        assert draws.get("theta") is not None

    @pytest.mark.skipif(pystan_version() != 2, reason="PyStan 2.x required")
    def test_index_order(self, data, eight_schools_params):
        """Test 0-indexed data."""
        # Skip test if pystan not installed
        pystan = importorskip("pystan")  # pylint: disable=import-error

        fit = data.model.sampling(data=eight_schools_params)
        if pystan.__version__ >= "2.18":
            # make 1-indexed to 0-indexed
            for holder in fit.sim["samples"]:
                new_chains = OrderedDict()
                for i, (key, values) in enumerate(holder.chains.items()):
                    if "[" in key:
                        name, *shape = key.replace("]", "").split("[")
                        shape = [str(int(item) - 1) for items in shape for item in items.split(",")]
                        key = f"{name}[{','.join(shape)}]"
                    new_chains[key] = np.full_like(values, fill_value=float(i))
                setattr(holder, "chains", new_chains)
            fit.sim["fnames_oi"] = list(fit.sim["samples"][0].chains.keys())
        idata = from_pystan(posterior=fit)
        assert idata is not None
        for j, fpar in enumerate(fit.sim["fnames_oi"]):
            par, *shape = fpar.replace("]", "").split("[")
            if par in {"lp__", "log_lik"}:
                continue
            assert hasattr(idata.posterior, par), (par, list(idata.posterior.data_vars))
            if shape:
                shape = [slice(None), slice(None)] + list(map(int, shape))
                assert idata.posterior[par][tuple(shape)].values.mean() == float(j)
            else:
                assert idata.posterior[par].values.mean() == float(j)



"""Bayesian p-value Posterior/Prior predictive plot."""

import numpy as np

from ..labels import BaseLabeller
from ..rcparams import rcParams
from ..utils import _var_names
from .plot_utils import default_grid, filter_plotters_list, get_plotting_function
from ..sel_utils import xarray_var_iter


def plot_bpv(
    data,
    kind="u_value",
    t_stat="median",
    bpv=True,
    plot_mean=True,
    reference="analytical",
    smoothing=None,
    mse=False,
    n_ref=100,
    hdi_prob=0.94,
    color="C0",
    grid=None,
    figsize=None,
    textsize=None,
    labeller=None,
    data_pairs=None,
    var_names=None,
    filter_vars=None,
    coords=None,
    flatten=None,
    flatten_pp=None,
    ax=None,
    backend=None,
    plot_ref_kwargs=None,
    backend_kwargs=None,
    group="posterior",
    show=None,
):
    r"""Plot Bayesian p-value for observed data and Posterior/Prior predictive.

    Parameters
    ----------
    data : InferenceData
        :class:`arviz.InferenceData` object containing the observed and
        posterior/prior predictive data.
    kind : {"u_value", "p_value", "t_stat"}, default "u_value"
        Specify the kind of plot:

        * The ``kind="p_value"`` computes :math:`p := p(y* \leq y | y)`.
          This is the probability of the data y being larger or equal than the predicted data y*.
          The ideal value is 0.5 (half the predictions below and half above the data).
        * The ``kind="u_value"`` argument computes :math:`p_i := p(y_i* \leq y_i | y)`.
          i.e. like a p_value but per observation :math:`y_i`. This is also known as marginal
          p_value. The ideal distribution is uniform. This is similar to the LOO-PIT
          calculation/plot, the difference is than in LOO-pit plot we compute
          :math:`pi = p(y_i* r \leq y_i | y_{-i} )`, where :math:`y_{-i}`,
          is all other data except :math:`y_i`.
        * The ``kind="t_stat"`` argument computes :math:`:= p(T(y)* \leq T(y) | y)`
          where T is any test statistic. See ``t_stat`` argument below for details
          of available options.

    t_stat : str, float, or callable, default "median"
        Test statistics to compute from the observations and predictive distributions.
        Allowed strings are “mean”, “median” or “std”. Alternative a quantile can be passed
        as a float (or str) in the interval (0, 1). Finally a user defined function is also
        acepted, see examples section for details.
    bpv : bool, default True
        If True add the Bayesian p_value to the legend when ``kind = t_stat``.
    plot_mean : bool, default True
        Whether or not to plot the mean test statistic.
    reference : {"analytical", "samples", None}, default "analytical"
        How to compute the distributions used as reference for ``kind=u_values``
        or ``kind=p_values``. Use `None` to not plot any reference.
    smoothing : bool, optional
        If True and the data has integer dtype, smooth the data before computing the p-values,
        u-values or tstat. By default, True when `kind` is "u_value" and False otherwise.
    mse : bool, default False
        Show scaled mean square error between uniform distribution and marginal p_value
        distribution.
    n_ref : int, default 100
        Number of reference distributions to sample when ``reference=samples``.
    hdi_prob : float, optional
        Probability for the highest density interval for the analytical reference distribution when
        ``kind=u_values``. Should be in the interval (0, 1]. Defaults to the
        rcParam ``stats.ci_prob``. See :ref:`this section <common_hdi_prob>` for usage examples.
    color : str, optional
        Matplotlib color
    grid : tuple, optional
        Number of rows and columns. By default, the rows and columns are
        automatically inferred. See :ref:`this section <common_grid>` for usage examples.
    figsize : (float, float), optional
        Figure size. If None it will be defined automatically.
    textsize : float, optional
        Text size scaling factor for labels, titles and lines. If None it will be autoscaled based
        on `figsize`.
    data_pairs : dict, optional
        Dictionary containing relations between observed data and posterior/prior predictive data.
        Dictionary structure:

        - key = data var_name
        - value = posterior/prior predictive var_name

        For example, ``data_pairs = {'y' : 'y_hat'}``
        If None, it will assume that the observed data and the posterior/prior
        predictive data have the same variable name.
    Labeller : Labeller, optional
        Class providing the method ``make_pp_label`` to generate the labels in the plot titles.
        Read the :ref:`label_guide` for more details and usage examples.
    var_names : list of str, optional
        Variables to be plotted. If `None` all variable are plotted. Prefix the variables by ``~``
        when you want to exclude them from the plot. See the :ref:`this section <common_var_names>`
        for usage examples. See :ref:`this section <common_var_names>` for usage examples.
    filter_vars : {None, "like", "regex"}, default None
        If `None` (default), interpret `var_names` as the real variables names. If "like",
        interpret `var_names` as substrings of the real variables names. If "regex",
        interpret `var_names` as regular expressions on the real variables names. See
        :ref:`this section <common_filter_vars>` for usage examples.
    coords : dict, optional
        Dictionary mapping dimensions to selected coordinates to be plotted.
        Dimensions without a mapping specified will include all coordinates for
        that dimension. Defaults to including all coordinates for all
        dimensions if None. See :ref:`this section <common_coords>` for usage examples.
    flatten : list, optional
        List of dimensions to flatten in observed_data. Only flattens across the coordinates
        specified in the coords argument. Defaults to flattening all of the dimensions.
    flatten_pp : list, optional
        List of dimensions to flatten in posterior_predictive/prior_predictive. Only flattens
        across the coordinates specified in the coords argument. Defaults to flattening all
        of the dimensions. Dimensions should match flatten excluding dimensions for data_pairs
        parameters. If `flatten` is defined and `flatten_pp` is None, then ``flatten_pp=flatten``.
    legend : bool, default True
        Add legend to figure.
    ax : 2D array-like of matplotlib_axes or bokeh_figure, optional
        A 2D array of locations into which to plot the densities. If not supplied, ArviZ will create
        its own array of plot areas (and return it).
    backend : str, optional
        Select plotting backend {"matplotlib", "bokeh"}. Default "matplotlib".
    plot_ref_kwargs :  dict, optional
        Extra keyword arguments to control how reference is represented.
        Passed to :meth:`matplotlib.axes.Axes.plot` or
        :meth:`matplotlib.axes.Axes.axhspan` (when ``kind=u_value``
        and ``reference=analytical``).
    backend_kwargs : bool, optional
        These are kwargs specific to the backend being used, passed to
        :func:`matplotlib.pyplot.subplots` or :class:`bokeh.plotting.figure`.
        For additional documentation check the plotting method of the backend.
    group : {"posterior", "prior"}, default "posterior"
        Specifies which InferenceData group should be plotted. If "posterior", then the values
        in `posterior_predictive` group are compared to the ones in `observed_data`, if "prior" then
        the same comparison happens, but with the values in `prior_predictive` group.
    show : bool, optional
        Call backend show function.

    Returns
    -------
    axes : 2D ndarray of matplotlib_axes or bokeh_figure

    See Also
    --------
    plot_ppc : Plot for posterior/prior predictive checks.
    plot_loo_pit : Plot Leave-One-Out probability integral transformation (PIT) predictive checks.
    plot_dist_comparison : Plot to compare fitted and unfitted distributions.

    References
    ----------
    * Gelman et al. (2013) see http://www.stat.columbia.edu/~gelman/book/ pages 151-153 for details

    Notes
    -----
    Discrete data is smoothed before computing either p-values or u-values using the
    function :func:`~arviz.smooth_data` if the data is integer type
    and the smoothing parameter is True.

    Examples
    --------
    Plot Bayesian p_values.

    .. plot::
        :context: close-figs

        >>> import arviz as az
        >>> data = az.load_arviz_data("regression1d")
        >>> az.plot_bpv(data, kind="p_value")

    Plot custom test statistic comparison.

    .. plot::
        :context: close-figs

        >>> import arviz as az
        >>> data = az.load_arviz_data("regression1d")
        >>> az.plot_bpv(data, kind="t_stat", t_stat=lambda x:np.percentile(x, q=50, axis=-1))
    """
    if group not in ("posterior", "prior"):
        raise TypeError("`group` argument must be either `posterior` or `prior`")

    for groups in (f"{group}_predictive", "observed_data"):
        if not hasattr(data, groups):
            raise TypeError(f'`data` argument must have the group "{groups}"')

    if kind.lower() not in ("t_stat", "u_value", "p_value"):
        raise TypeError("`kind` argument must be either `t_stat`, `u_value`, or `p_value`")

    if reference is not None and reference.lower() not in ("analytical", "samples"):
        raise TypeError("`reference` argument must be either `analytical`, `samples`, or `None`")

    if hdi_prob is None:
        hdi_prob = rcParams["stats.ci_prob"]
    elif not 1 >= hdi_prob > 0:
        raise ValueError("The value of hdi_prob should be in the interval (0, 1]")

    if smoothing is None:
        smoothing = kind.lower() == "u_value"

    if data_pairs is None:
        data_pairs = {}

    if labeller is None:
        labeller = BaseLabeller()

    if backend is None:
        backend = rcParams["plot.backend"]
    backend = backend.lower()

    observed = data.observed_data

    if group == "posterior":
        predictive_dataset = data.posterior_predictive
    elif group == "prior":
        predictive_dataset = data.prior_predictive

    if var_names is None:
        var_names = list(observed.data_vars)
    var_names = _var_names(var_names, observed, filter_vars)
    pp_var_names = [data_pairs.get(var, var) for var in var_names]
    pp_var_names = _var_names(pp_var_names, predictive_dataset, filter_vars)

    if flatten_pp is None:
        if flatten is None:
            flatten_pp = list(predictive_dataset.dims)
        else:
            flatten_pp = flatten
    if flatten is None:
        flatten = list(observed.dims)

    if coords is None:
        coords = {}

    total_pp_samples = predictive_dataset.sizes["chain"] * predictive_dataset.sizes["draw"]

    for key in coords.keys():
        coords[key] = np.where(np.isin(observed[key], coords[key]))[0]

    obs_plotters = filter_plotters_list(
        list(
            xarray_var_iter(
                observed.isel(coords), skip_dims=set(flatten), var_names=var_names, combined=True
            )
        ),
        "plot_t_stats",
    )
    length_plotters = len(obs_plotters)
    pp_plotters = [
        tup
        for _, tup in zip(
            range(length_plotters),
            xarray_var_iter(
                predictive_dataset.isel(coords),
                var_names=pp_var_names,
                skip_dims=set(flatten_pp),
                combined=True,
            ),
        )
    ]
    rows, cols = default_grid(length_plotters, grid=grid)

    bpvplot_kwargs = dict(
        ax=ax,
        length_plotters=length_plotters,
        rows=rows,
        cols=cols,
        obs_plotters=obs_plotters,
        pp_plotters=pp_plotters,
        total_pp_samples=total_pp_samples,
        kind=kind,
        bpv=bpv,
        t_stat=t_stat,
        reference=reference,
        mse=mse,
        n_ref=n_ref,
        hdi_prob=hdi_prob,
        plot_mean=plot_mean,
        color=color,
        figsize=figsize,
        textsize=textsize,
        labeller=labeller,
        plot_ref_kwargs=plot_ref_kwargs,
        backend_kwargs=backend_kwargs,
        show=show,
        smoothing=smoothing,
    )

    # TODO: Add backend kwargs
    plot = get_plotting_function("plot_bpv", "bpvplot", backend)
    axes = plot(**bpvplot_kwargs)
    return axes



# Plotting and reporting Bayes Factor given idata, var name, prior distribution and reference value
# pylint: disable=unbalanced-tuple-unpacking
import logging

from ..data.utils import extract
from .plot_utils import get_plotting_function
from ..stats import bayes_factor

_log = logging.getLogger(__name__)


def plot_bf(
    idata,
    var_name,
    prior=None,
    ref_val=0,
    colors=("C0", "C1"),
    figsize=None,
    textsize=None,
    hist_kwargs=None,
    plot_kwargs=None,
    ax=None,
    backend=None,
    backend_kwargs=None,
    show=None,
):
    r"""Approximated Bayes Factor for comparing hypothesis of two nested models.

    The Bayes factor is estimated by comparing a model (H1) against a model in which the
    parameter of interest has been restricted to be a point-null (H0). This computation
    assumes the models are nested and thus H0 is a special case of H1.

    Notes
    -----
    The bayes Factor is approximated as the Savage-Dickey density ratio
    algorithm presented in [1]_.

    Parameters
    ----------
    idata : InferenceData
        Any object that can be converted to an :class:`arviz.InferenceData` object
        Refer to documentation of :func:`arviz.convert_to_dataset` for details.
    var_name : str, optional
        Name of variable we want to test.
    prior : numpy.array, optional
        In case we want to use different prior, for example for sensitivity analysis.
    ref_val : int, default 0
        Point-null for Bayes factor estimation.
    colors : tuple, default ('C0', 'C1')
        Tuple of valid Matplotlib colors. First element for the prior, second for the posterior.
    figsize : (float, float), optional
        Figure size. If `None` it will be defined automatically.
    textsize : float, optional
        Text size scaling factor for labels, titles and lines. If `None` it will be auto
        scaled based on `figsize`.
    plot_kwargs : dict, optional
        Additional keywords passed to :func:`matplotlib.pyplot.plot`.
    hist_kwargs : dict, optional
        Additional keywords passed to :func:`arviz.plot_dist`. Only works for discrete variables.
    ax : axes, optional
        :class:`matplotlib.axes.Axes` or :class:`bokeh.plotting.Figure`.
    backend : {"matplotlib", "bokeh"}, default "matplotlib"
        Select plotting backend.
    backend_kwargs : dict, optional
        These are kwargs specific to the backend being used, passed to
        :func:`matplotlib.pyplot.subplots` or :class:`bokeh.plotting.figure`.
        For additional documentation check the plotting method of the backend.
    show : bool, optional
        Call backend show function.

    Returns
    -------
    dict : A dictionary with BF10 (Bayes Factor 10 (H1/H0 ratio), and BF01 (H0/H1 ratio).
    axes : matplotlib_axes or bokeh_figure

    References
    ----------
    .. [1] Heck, D., 2019. A caveat on the Savage-Dickey density ratio:
       The case of computing Bayes factors for regression parameters.

    Examples
    --------
    Moderate evidence indicating that the parameter "a" is different from zero.

    .. plot::
        :context: close-figs

        >>> import numpy as np
        >>> import arviz as az
        >>> idata = az.from_dict(posterior={"a":np.random.normal(1, 0.5, 5000)},
        ...     prior={"a":np.random.normal(0, 1, 5000)})
        >>> az.plot_bf(idata, var_name="a", ref_val=0)

    """

    if prior is None:
        prior = extract(idata, var_names=var_name, group="prior").values

    bf, p_at_ref_val = bayes_factor(
        idata, var_name, prior=prior, ref_val=ref_val, return_ref_vals=True
    )
    bf_10 = bf["BF10"]
    bf_01 = bf["BF01"]

    posterior = extract(idata, var_names=var_name)

    bfplot_kwargs = dict(
        ax=ax,
        bf_10=bf_10.item(),
        bf_01=bf_01.item(),
        prior=prior,
        posterior=posterior,
        ref_val=ref_val,
        prior_at_ref_val=p_at_ref_val["prior"],
        posterior_at_ref_val=p_at_ref_val["posterior"],
        var_name=var_name,
        colors=colors,
        figsize=figsize,
        textsize=textsize,
        hist_kwargs=hist_kwargs,
        plot_kwargs=plot_kwargs,
        backend_kwargs=backend_kwargs,
        show=show,
    )

    plot = get_plotting_function("plot_bf", "bfplot", backend)
    axes = plot(**bfplot_kwargs)
    return {"BF10": bf_10, "BF01": bf_01}, axes



"""Plot timeseries data."""

import warnings
import numpy as np

from ..sel_utils import xarray_var_iter
from ..rcparams import rcParams
from .plot_utils import default_grid, get_plotting_function


def plot_ts(
    idata,
    y,
    x=None,
    y_hat=None,
    y_holdout=None,
    y_forecasts=None,
    x_holdout=None,
    plot_dim=None,
    holdout_dim=None,
    num_samples=100,
    backend=None,
    backend_kwargs=None,
    y_kwargs=None,
    y_hat_plot_kwargs=None,
    y_mean_plot_kwargs=None,
    vline_kwargs=None,
    textsize=None,
    figsize=None,
    legend=True,
    axes=None,
    show=None,
):
    """Plot timeseries data.

    Parameters
    ----------
    idata : InferenceData
        :class:`arviz.InferenceData` object.
    y : str
        Variable name from ``observed_data``.
        Values to be plotted on y-axis before holdout.
    x : str, Optional
        Values to be plotted on x-axis before holdout.
        If None, coords of ``y`` dims is chosen.
    y_hat : str, optional
        Variable name from ``posterior_predictive``.
        Assumed to be of shape ``(chain, draw, *y_dims)``.
    y_holdout : str, optional
        Variable name from ``observed_data``.
        It represents the observed data after the holdout period.
        Useful while testing the model, when you want to compare
        observed test data with predictions/forecasts.
    y_forecasts : str, optional
        Variable name from ``posterior_predictive``.
        It represents forecasts (posterior predictive) values after holdout period.
        Useful to compare observed vs predictions/forecasts.
        Assumed shape ``(chain, draw, *shape)``.
    x_holdout : str, Defaults to coords of y.
        Variable name from ``constant_data``.
        If None, coords of ``y_holdout`` or
        coords of ``y_forecast`` (either of the two available) is chosen.
    plot_dim: str, Optional
        Should be present in ``y.dims``.
        Necessary for selection of ``x`` if ``x`` is None and ``y`` is multidimensional.
    holdout_dim: str, Optional
        Should be present in ``y_holdout.dims`` or ``y_forecats.dims``.
        Necessary to choose ``x_holdout`` if ``x`` is None and
        if ``y_holdout`` or ``y_forecasts`` is multidimensional.
    num_samples : int, default 100
        Number of posterior predictive samples drawn from ``y_hat`` and ``y_forecasts``.
    backend : {"matplotlib", "bokeh"}, default "matplotlib"
        Select plotting backend.
    y_kwargs : dict, optional
        Passed to :meth:`matplotlib.axes.Axes.plot` in matplotlib.
    y_hat_plot_kwargs : dict, optional
        Passed to :meth:`matplotlib.axes.Axes.plot` in matplotlib.
    y_mean_plot_kwargs : dict, optional
        Passed to :meth:`matplotlib.axes.Axes.plot` in matplotlib.
    vline_kwargs : dict, optional
        Passed to :meth:`matplotlib.axes.Axes.axvline` in matplotlib.
    backend_kwargs : dict, optional
        These are kwargs specific to the backend being used. Passed to
        :func:`matplotlib.pyplot.subplots`.
    figsize : tuple, optional
        Figure size. If None, it will be defined automatically.
    textsize : float, optional
        Text size scaling factor for labels, titles and lines. If None, it will be
        autoscaled based on ``figsize``.


    Returns
    -------
    axes: matplotlib axes or bokeh figures.

    See Also
    --------
    plot_lm : Posterior predictive and mean plots for regression-like data.
    plot_ppc : Plot for posterior/prior predictive checks.

    Examples
    --------
    Plot timeseries default plot

    .. plot::
        :context: close-figs

        >>> import arviz as az
        >>> nchains, ndraws = (4, 500)
        >>> obs_data = {
        ...     "y": 2 * np.arange(1, 9) + 3,
        ...     "z": 2 * np.arange(8, 12) + 3,
        ... }
        >>> posterior_predictive = {
        ...     "y": np.random.normal(
        ...         (obs_data["y"] * 1.2) - 3, size=(nchains, ndraws, len(obs_data["y"]))
        ...     ),
        ...     "z": np.random.normal(
        ...         (obs_data["z"] * 1.2) - 3, size=(nchains, ndraws, len(obs_data["z"]))
        ...     ),
        ...  }
        >>> idata = az.from_dict(
        ...     observed_data=obs_data,
        ...     posterior_predictive=posterior_predictive,
        ...     coords={"obs_dim": np.arange(1, 9), "pred_dim": np.arange(8, 12)},
        ...     dims={"y": ["obs_dim"], "z": ["pred_dim"]},
        ... )
        >>> ax = az.plot_ts(idata=idata, y="y", y_holdout="z")

    Plot timeseries multidim plot

    .. plot::
        :context: close-figs

        >>> ndim1, ndim2 = (5, 7)
        >>> data = {
        ...     "y": np.random.normal(size=(ndim1, ndim2)),
        ...     "z": np.random.normal(size=(ndim1, ndim2)),
        ... }
        >>> posterior_predictive = {
        ...     "y": np.random.randn(nchains, ndraws, ndim1, ndim2),
        ...     "z": np.random.randn(nchains, ndraws, ndim1, ndim2),
        ... }
        >>> const_data = {"x": np.arange(1, 6), "x_pred": np.arange(5, 10)}
        >>> idata = az.from_dict(
        ...     observed_data=data,
        ...     posterior_predictive=posterior_predictive,
        ...     constant_data=const_data,
        ...     dims={
        ...         "y": ["dim1", "dim2"],
        ...         "z": ["holdout_dim1", "holdout_dim2"],
        ...     },
        ...     coords={
        ...         "dim1": range(ndim1),
        ...         "dim2": range(ndim2),
        ...         "holdout_dim1": range(ndim1 - 1, ndim1 + 4),
        ...         "holdout_dim2": range(ndim2 - 1, ndim2 + 6),
        ...     },
        ... )
        >>> az.plot_ts(
        ...     idata=idata,
        ...     y="y",
        ...     plot_dim="dim1",
        ...     y_holdout="z",
        ...     holdout_dim="holdout_dim1",
        ... )

    """
    # Assign default values if none is provided
    y_hat = y if y_hat is None and isinstance(y, str) else y_hat
    y_forecasts = y_holdout if y_forecasts is None and isinstance(y_holdout, str) else y_forecasts
    # holdout_dim = plot_dim if holdout_dim is None and plot_dim is not None else holdout_dim

    if isinstance(y, str):
        y = idata.observed_data[y]

    if isinstance(y_holdout, str):
        y_holdout = idata.observed_data[y_holdout]

    if len(y.dims) > 1 and plot_dim is None:
        raise ValueError("Argument plot_dim is needed in case of multidimensional data")

    if y_holdout is not None and len(y_holdout.dims) > 1 and holdout_dim is None:
        raise ValueError("Argument holdout_dim is needed in case of multidimensional data")

    # Assigning values to x
    x_var_names = None
    if isinstance(x, str):
        x = idata.constant_data[x]
    elif isinstance(x, tuple):
        x_var_names = x
        x = idata.constant_data
    elif x is None:
        if plot_dim is None:
            x = y.coords[y.dims[0]]
        else:
            x = y.coords[plot_dim]

    # If posterior_predictive is present in idata and y_hat is there, get its values
    if isinstance(y_hat, str):
        if "posterior_predictive" not in idata.groups():
            warnings.warn("posterior_predictive not found in idata", UserWarning)
            y_hat = None
        elif hasattr(idata.posterior_predictive, y_hat):
            y_hat = idata.posterior_predictive[y_hat]
        else:
            warnings.warn("y_hat not found in posterior_predictive", UserWarning)
            y_hat = None

    # If posterior_predictive is present in idata and y_forecasts is there, get its values
    x_holdout_var_names = None
    if isinstance(y_forecasts, str):
        if "posterior_predictive" not in idata.groups():
            warnings.warn("posterior_predictive not found in idata", UserWarning)
            y_forecasts = None
        elif hasattr(idata.posterior_predictive, y_forecasts):
            y_forecasts = idata.posterior_predictive[y_forecasts]
        else:
            warnings.warn("y_hat not found in posterior_predictive", UserWarning)
            y_forecasts = None

    # Assign values to y_holdout
    if isinstance(y_holdout, str):
        y_holdout = idata.observed_data[y_holdout]

    # Assign values to x_holdout.
    if y_holdout is not None or y_forecasts is not None:
        if x_holdout is None:
            if holdout_dim is None:
                if y_holdout is None:
                    x_holdout = y_forecasts.coords[y_forecasts.dims[-1]]
                else:
                    x_holdout = y_holdout.coords[y_holdout.dims[-1]]
            elif y_holdout is None:
                x_holdout = y_forecasts.coords[holdout_dim]
            else:
                x_holdout = y_holdout.coords[holdout_dim]
        elif isinstance(x_holdout, str):
            x_holdout = idata.constant_data[x_holdout]
        elif isinstance(x_holdout, tuple):
            x_holdout_var_names = x_holdout
            x_holdout = idata.constant_data

    # Choose dims to generate y plotters
    if plot_dim is None:
        skip_dims = list(y.dims)
    elif isinstance(plot_dim, str):
        skip_dims = [plot_dim]
    elif isinstance(plot_dim, tuple):
        skip_dims = list(plot_dim)

    # Choose dims to generate y_holdout plotters
    if holdout_dim is None:
        if y_holdout is not None:
            skip_holdout_dims = list(y_holdout.dims)
        elif y_forecasts is not None:
            skip_holdout_dims = list(y_forecasts.dims)
    elif isinstance(holdout_dim, str):
        skip_holdout_dims = [holdout_dim]
    elif isinstance(holdout_dim, tuple):
        skip_holdout_dims = list(holdout_dim)

    # Compulsory plotters
    y_plotters = list(
        xarray_var_iter(
            y,
            skip_dims=set(skip_dims),
            combined=True,
        )
    )

    # Compulsory plotters
    x_plotters = list(
        xarray_var_iter(
            x,
            var_names=x_var_names,
            skip_dims=set(x.dims),
            combined=True,
        )
    )
    # Necessary when multidim y
    # If there are multiple x and multidimensional y, we need total of len(x)*len(y) graphs
    len_y = len(y_plotters)
    len_x = len(x_plotters)
    length_plotters = len_x * len_y
    # TODO: Incompatible types in assignment (expression has type "ndarray[Any, dtype[Any]]",
    # TODO: variable has type "List[Any]")  [assignment]
    y_plotters = np.tile(np.array(y_plotters, dtype=object), (len_x, 1))  # type: ignore[assignment]
    x_plotters = np.tile(np.array(x_plotters, dtype=object), (len_y, 1))  # type: ignore[assignment]

    # Generate plotters for all the available data
    y_mean_plotters = None
    y_hat_plotters = None
    if y_hat is not None:
        total_samples = y_hat.sizes["chain"] * y_hat.sizes["draw"]
        pp_sample_ix = np.random.choice(total_samples, size=num_samples, replace=False)

        y_hat_satcked = y_hat.stack(__sample__=("chain", "draw"))[..., pp_sample_ix]

        y_hat_plotters = list(
            xarray_var_iter(
                y_hat_satcked,
                skip_dims=set(skip_dims + ["__sample__"]),
                combined=True,
            )
        )

        y_mean = y_hat.mean(("chain", "draw"))
        y_mean_plotters = list(
            xarray_var_iter(
                y_mean,
                skip_dims=set(skip_dims),
                combined=True,
            )
        )

        # Necessary when multidim y
        # If there are multiple x and multidimensional y, we need total of len(x)*len(y) graphs
        # TODO: Incompatible types in assignment (expression has type "ndarray[Any, dtype[Any]]",
        # TODO: variable has type "List[Any]")  [assignment]
        y_hat_plotters = np.tile(
            np.array(y_hat_plotters, dtype=object), (len_x, 1)
        )  # type: ignore[assignment]
        y_mean_plotters = np.tile(
            np.array(y_mean_plotters, dtype=object), (len_x, 1)
        )  # type: ignore[assignment]

    y_holdout_plotters = None
    x_holdout_plotters = None
    if y_holdout is not None:
        y_holdout_plotters = list(
            xarray_var_iter(
                y_holdout,
                skip_dims=set(skip_holdout_dims),
                combined=True,
            )
        )

        x_holdout_plotters = list(
            xarray_var_iter(
                x_holdout,
                var_names=x_holdout_var_names,
                skip_dims=set(x_holdout.dims),
                combined=True,
            )
        )

        # Necessary when multidim y
        # If there are multiple x and multidimensional y, we need total of len(x)*len(y) graphs
        # TODO: Incompatible types in assignment (expression has type "ndarray[Any, dtype[Any]]",
        # TODO: variable has type "List[Any]")  [assignment]
        y_holdout_plotters = np.tile(
            np.array(y_holdout_plotters, dtype=object), (len_x, 1)
        )  # type: ignore[assignment]
        x_holdout_plotters = np.tile(
            np.array(x_holdout_plotters, dtype=object), (len_y, 1)
        )  # type: ignore[assignment]

    y_forecasts_plotters = None
    y_forecasts_mean_plotters = None
    if y_forecasts is not None:
        total_samples = y_forecasts.sizes["chain"] * y_forecasts.sizes["draw"]
        pp_sample_ix = np.random.choice(total_samples, size=num_samples, replace=False)

        y_forecasts_satcked = y_forecasts.stack(__sample__=("chain", "draw"))[..., pp_sample_ix]

        y_forecasts_plotters = list(
            xarray_var_iter(
                y_forecasts_satcked,
                skip_dims=set(skip_holdout_dims + ["__sample__"]),
                combined=True,
            )
        )

        y_forecasts_mean = y_forecasts.mean(("chain", "draw"))
        y_forecasts_mean_plotters = list(
            xarray_var_iter(
                y_forecasts_mean,
                skip_dims=set(skip_holdout_dims),
                combined=True,
            )
        )

        x_holdout_plotters = list(
            xarray_var_iter(
                x_holdout,
                var_names=x_holdout_var_names,
                skip_dims=set(x_holdout.dims),
                combined=True,
            )
        )

        # Necessary when multidim y
        # If there are multiple x and multidimensional y, we need total of len(x)*len(y) graphs
        # TODO: Incompatible types in assignment (expression has type "ndarray[Any, dtype[Any]]",
        # TODO: variable has type "List[Any]")  [assignment]
        y_forecasts_mean_plotters = np.tile(
            np.array(y_forecasts_mean_plotters, dtype=object), (len_x, 1)
        )  # type: ignore[assignment]
        y_forecasts_plotters = np.tile(
            np.array(y_forecasts_plotters, dtype=object), (len_x, 1)
        )  # type: ignore[assignment]
        x_holdout_plotters = np.tile(
            np.array(x_holdout_plotters, dtype=object), (len_y, 1)
        )  # type: ignore[assignment]

    rows, cols = default_grid(length_plotters)

    tsplot_kwargs = dict(
        x_plotters=x_plotters,
        y_plotters=y_plotters,
        y_mean_plotters=y_mean_plotters,
        y_hat_plotters=y_hat_plotters,
        y_holdout_plotters=y_holdout_plotters,
        x_holdout_plotters=x_holdout_plotters,
        y_forecasts_plotters=y_forecasts_plotters,
        y_forecasts_mean_plotters=y_forecasts_mean_plotters,
        num_samples=num_samples,
        length_plotters=length_plotters,
        rows=rows,
        cols=cols,
        backend_kwargs=backend_kwargs,
        y_kwargs=y_kwargs,
        y_hat_plot_kwargs=y_hat_plot_kwargs,
        y_mean_plot_kwargs=y_mean_plot_kwargs,
        vline_kwargs=vline_kwargs,
        textsize=textsize,
        figsize=figsize,
        legend=legend,
        axes=axes,
        show=show,
    )

    if backend is None:
        backend = rcParams["plot.backend"]
    backend = backend.lower()

    plot = get_plotting_function("plot_ts", "tsplot", backend)
    ax = plot(**tsplot_kwargs)
    return ax



"""Matplotlib Parallel coordinates plot."""

import matplotlib.pyplot as plt
import numpy as np

from ...plot_utils import _scale_fig_size
from . import backend_kwarg_defaults, backend_show, create_axes_grid


def plot_parallel(
    ax,
    colornd,
    colord,
    shadend,
    diverging_mask,
    posterior,
    textsize,
    var_names,
    legend,
    figsize,
    backend_kwargs,
    backend_config,  # pylint: disable=unused-argument
    show,
):
    """Matplotlib parallel plot."""
    if backend_kwargs is None:
        backend_kwargs = {}

    backend_kwargs = {
        **backend_kwarg_defaults(),
        **backend_kwargs,
    }

    figsize, _, _, xt_labelsize, _, _ = _scale_fig_size(figsize, textsize, 1, 1)
    backend_kwargs.setdefault("figsize", figsize)
    backend_kwargs["squeeze"] = True
    if ax is None:
        _, ax = create_axes_grid(1, backend_kwargs=backend_kwargs)

    ax.plot(posterior[:, ~diverging_mask], color=colornd, alpha=shadend)

    if np.any(diverging_mask):
        ax.plot(posterior[:, diverging_mask], color=colord, lw=1)

    ax.tick_params(labelsize=textsize)
    ax.set_xticks(range(len(var_names)))
    ax.set_xticklabels(var_names)

    if legend:
        ax.plot([], color=colornd, label="non-divergent")
        if np.any(diverging_mask):
            ax.plot([], color=colord, label="divergent")
        ax.legend(fontsize=xt_labelsize)

    if backend_show(show):
        plt.show()

    return ax
